# -*- coding: utf-8 -*-
"""
It is better to use constant variables instead of hoping you spell the same
string correctly every time you use it. (Also it makes it much easier if a
string name changes)
"""
from __future__ import absolute_import, division, print_function, unicode_literals
import six
import numpy as np
import math
import utool as ut
from collections import OrderedDict
from os.path import join
ut.noinject('[const]')


PI  = math.pi
TAU = 2.0 * PI

VIEWTEXT_TO_YAW_RADIANS = OrderedDict([
    ('right'      , 0.000 * TAU,),
    ('frontright' , 0.125 * TAU,),
    ('front'      , 0.250 * TAU,),
    ('frontleft'  , 0.375 * TAU,),
    ('left'       , 0.500 * TAU,),
    ('backleft'   , 0.625 * TAU,),
    ('back'       , 0.750 * TAU,),
    ('backright'  , 0.875 * TAU,),
])

# Mapping of viewpoints codes to yaw angles
VIEWTEXT_TO_YAW_RADIANS = OrderedDict([
    ('right'      , 0.000 * TAU,),
    ('frontright' , 0.125 * TAU,),
    ('front'      , 0.250 * TAU,),
    ('frontleft'  , 0.375 * TAU,),
    ('left'       , 0.500 * TAU,),
    ('backleft'   , 0.625 * TAU,),
    ('back'       , 0.750 * TAU,),
    ('backright'  , 0.875 * TAU,),
])
VIEWTEXT_TO_VIEWPOINT_RADIANS = VIEWTEXT_TO_YAW_RADIANS


# TODO: DEPRICATE ALL THE YAW ALIAS STUFF IN FAVOR OF const.VIEW
YAWALIAS = {
    # maybe this can be CODE_TO_SHORT
    'up':             'U',
    'down':           'D',
    'front':          'F',
    'left':           'L',
    'back':           'B',
    'right':          'R',
    'upfront':        'UF',
    'upback':         'UB',
    'upleft':         'UL',
    'upright':        'UR',
    'downfront':      'DF',
    'downback':       'DB',
    'downleft':       'DL',
    'downright':      'DR',
    'frontleft':      'FL',
    'frontright':     'FR',
    'backleft':       'BL',
    'backright':      'BR',
    'upfrontleft':    'UFL',
    'upfrontright':   'UFR',
    'upbackleft':     'UBL',
    'upbackright':    'UBR',
    'downfrontleft':  'DFL',
    'downfrontright': 'DFR',
    'downbackleft':   'DBL',
    'downbackright':  'DBR',
}

QUAL_EXCELLENT = 'excellent'
QUAL_GOOD      = 'good'
QUAL_OK        = 'ok'
QUAL_POOR      = 'poor'
QUAL_JUNK      = 'junk'
QUAL_UNKNOWN   = 'UNKNOWN'

QUALITY_INT_TO_TEXT = OrderedDict([
    (5,  QUAL_EXCELLENT,),
    (4,  QUAL_GOOD,),
    (3,  QUAL_OK,),
    (2,  QUAL_POOR,),
    # oops forgot 1. will be mapped to poor
    (0,  QUAL_JUNK,),
    (-1, QUAL_UNKNOWN,),
])


QUALITY_TEXT_TO_INT       = ut.invert_dict(QUALITY_INT_TO_TEXT)
QUALITY_INT_TO_TEXT[1]    = QUAL_JUNK
#QUALITY_TEXT_TO_INTS      = ut.invert_dict(QUALITY_INT_TO_TEXT)
QUALITY_TEXT_TO_INTS = ut.group_items(
    list(QUALITY_INT_TO_TEXT.keys()),
    list(QUALITY_INT_TO_TEXT.values()))
QUALITY_TEXT_TO_INTS[QUAL_UNKNOWN] = -1
QUALITY_INT_TO_TEXT[None] = QUALITY_INT_TO_TEXT[-1]


SEX_INT_TO_TEXT = {
    None: 'UNKNOWN NAME',
    -1  : 'UNKNOWN SEX',
    0   : 'Female',
    1   : 'Male',
    2   : 'INDETERMINATE SEX',
}
SEX_TEXT_TO_INT = ut.invert_dict(SEX_INT_TO_TEXT)


class PATH_NAMES(object):
    """ Path names for internal IBEIS database """
    sqldb       = '_ibeis_database.sqlite3'
    sqlstaging  = '_ibeis_staging.sqlite3'
    _ibsdb      = '_ibsdb'
    cache       = '_ibeis_cache'
    backups     = '_ibeis_backups'
    logs        = '_ibeis_logs'
    chips       = 'chips'
    figures     = 'figures'
    flann       = 'flann'
    images      = 'images'
    trees       = 'trees'
    nets        = 'nets'
    uploads     = 'uploads'
    detectimg   = 'detectimg'
    thumbs      = 'thumbs'
    trashdir    = 'trashed_images'
    distinctdir = 'distinctiveness_model'
    scorenormdir = 'scorenorm'
    smartpatrol = 'smart_patrol'
    # Query Results (chipmatch dirs)
    qres       = 'qres_new'
    bigcache   = 'qres_bigcache_new'


class REL_PATHS(object):
    """ all paths are relative to ibs.dbdir """
    _ibsdb      = PATH_NAMES._ibsdb
    trashdir    = PATH_NAMES.trashdir
    figures     = join(_ibsdb, PATH_NAMES.figures)
    cache       = join(_ibsdb, PATH_NAMES.cache)
    backups     = join(_ibsdb, PATH_NAMES.backups)
    logs        = join(_ibsdb, PATH_NAMES.logs)
    #chips       = join(_ibsdb, PATH_NAMES.chips)
    images      = join(_ibsdb, PATH_NAMES.images)
    trees       = join(_ibsdb, PATH_NAMES.trees)
    nets        = join(_ibsdb, PATH_NAMES.nets)
    uploads     = join(_ibsdb, PATH_NAMES.uploads)
    # All computed dirs live in <dbdir>/_ibsdb/_ibeis_cache
    chips       = join(cache, PATH_NAMES.chips)
    thumbs      = join(cache, PATH_NAMES.thumbs)
    flann       = join(cache, PATH_NAMES.flann)
    qres        = join(cache, PATH_NAMES.qres)
    bigcache    = join(cache, PATH_NAMES.bigcache)
    distinctdir = join(cache, PATH_NAMES.distinctdir)


# Directories that should be excluded from copy operations
EXCLUDE_COPY_REL_DIRS = [
    REL_PATHS.chips,
    REL_PATHS.cache,
    REL_PATHS.backups,
    REL_PATHS.figures,
    REL_PATHS.nets,
    join(PATH_NAMES._ibsdb, '_ibeis_cache*'),
    #'_ibsdb/_ibeis_cache',
    '_ibsdb/chips',  # old path for caches
    './images',  # the hotspotter images dir
]


# TODO: Remove anything under this block completely


UNKNOWN_LBLANNOT_ROWID = 0
UNKNOWN_NAME_ROWID = 0
UNKNOWN_SPECIES_ROWID = 0
# Names normalized to the standard UNKNOWN_NAME
ACCEPTED_UNKNOWN_NAMES = set(['Unassigned'])

INDIVIDUAL_KEY = 'INDIVIDUAL_KEY'
SPECIES_KEY    = 'SPECIES_KEY'
EMPTY_KEY      = ''
UNKNOWN        = '____'
KEY_DEFAULTS   = {
    INDIVIDUAL_KEY : UNKNOWN,
    SPECIES_KEY    : UNKNOWN,
}

# <UNFINISHED METADATA>
# We are letting wildbook do this metadata instead
# Define the special metadata for annotation

ROSEMARY_ANNOT_METADATA = [
    ('local_name'    , 'Local name:',    str),
    ('sun'           , 'Sun:',           ['FS', 'PS', 'NS']),
    ('wind'          , 'Wind:',          ['NW', 'LW', 'MW', 'SW']),
    ('rain'          , 'Rain:',          ['NR', 'LR', 'MR', 'HR']),
    ('cover'         , 'Cover:',         float),
    ('grass'         , 'Grass:',         ['less hf', 'less hk', 'less belly']),
    ('grass_color'   , 'Grass Colour:',  ['B', 'BG', 'GB', 'G']),
    ('grass_species' , 'Grass Species:', str),
    ('bush_type'     , 'Bush type:',     ['OG', 'LB', 'MB', 'TB']),
    ('bit'           , 'Bit:',           int),
    ('other_speceis' , 'Other Species:', str),
]

#ROSEMARY_KEYS = utool.get_list_column(ROSEMARY_ANNOT_METADATA, 0)
#KEY_DEFAULTS.update(**{key: UNKNOWN for key in ROSEMARY_KEYS})
# </UNFINISHED METADATA>

BASE_DATABASE_VERSION = '0.0.0'

#################################################################
# DO NOT DELETE FROM THE TABLE LIST, THE DATABASE UPDATER WILL BREAK!!!
# THIS GOES FOR OLD AND DEPRICATED TABLENAMES AS WELL!!!
# TODO:
# What should happen is when they are depricated they should go into a
# depricated tablename structure with the relevant versions suffixed
#################################################################
# CORE DATABASE
AL_RELATION_TABLE    = 'annotation_lblannot_relationship'
GA_RELATION_TABLE    = 'annotgroup_annotation_relationship'
ANNOTGROUP_TABLE     = 'annotgroups'
ANNOTATION_TABLE     = 'annotations'
PART_TABLE           = 'parts'
CHIP_TABLE           = 'chips'
CONFIG_TABLE         = 'configs'
CONTRIBUTOR_TABLE    = 'contributors'
GSG_RELATION_TABLE   = 'imageset_image_relationship'
IMAGESET_TABLE       = 'imagesets'
FEATURE_TABLE        = 'features'
FEATURE_WEIGHT_TABLE = 'feature_weights'
GL_RELATION_TABLE    = 'image_lblimage_relationship'
IMAGE_TABLE          = 'images'
LBLANNOT_TABLE       = 'lblannot'
LBLIMAGE_TABLE       = 'lblimage'
LBLTYPE_TABLE        = 'keys'
METADATA_TABLE       = 'metadata'
# Ugly move from name to names, need better way of versioning old table names
NAME_TABLE_v121      = 'name'
NAME_TABLE_v130      = 'names'
NAME_TABLE           = NAME_TABLE_v130
ANNOTMATCH_TABLE     = 'annotmatch'
SPECIES_TABLE        = 'species'
RESIDUAL_TABLE       = 'residuals'
VERSIONS_TABLE       = 'versions'
#
PARTY_CONTRIB_RELATION_TABLE = 'party_contrib_relation'
PARTY_TABLE                  = 'party'
#################################################################
# STAGING DATABASE
REVIEW_TABLE          = 'reviews'
TEST_TABLE            = 'tests'
#################################################################


# DEPCACHE TABLENAMES
#CHIPTHUMB_TABLE = 'chipthumb'


UNKNOWN_PURPLE_RGBA255 = np.array((102,   0, 153, 255))
NAME_BLUE_RGBA255      = np.array((20, 20, 235, 255))
NAME_RED_RGBA255       = np.array((235, 20, 20, 255))
NEW_YELLOW_RGBA255     = np.array((235, 235, 20, 255))

UNKNOWN_PURPLE_RGBA01 = UNKNOWN_PURPLE_RGBA255 / 255.0
NAME_BLUE_RGBA01      = NAME_BLUE_RGBA255 / 255.0
NAME_RED_RGBA01       = NAME_RED_RGBA255 / 255.0
NEW_YELLOW_RGBA01     = NEW_YELLOW_RGBA255 / 255.0

EXEMPLAR_IMAGESETTEXT         = '*Exemplars'
ALL_IMAGE_IMAGESETTEXT        = '*All Images'
UNREVIEWED_IMAGE_IMAGESETTEXT = '*Undetected Images'
REVIEWED_IMAGE_IMAGESETTEXT   = '*Reviewed Detections'
UNGROUPED_IMAGES_IMAGESETTEXT = '*Ungrouped Images'
SPECIAL_IMAGESET_LABELS = [
    EXEMPLAR_IMAGESETTEXT,
    ALL_IMAGE_IMAGESETTEXT,
    UNREVIEWED_IMAGE_IMAGESETTEXT,
    REVIEWED_IMAGE_IMAGESETTEXT,
    UNGROUPED_IMAGES_IMAGESETTEXT
]
NEW_IMAGESET_IMAGESETTEXT = 'NEW IMAGESET'

#IMAGE_THUMB_SUFFIX = '_thumb.png'
#CHIP_THUMB_SUFFIX  = '_chip_thumb.png'
IMAGE_THUMB_SUFFIX = '_thumb.jpg'
IMAGE_BARE_THUMB_SUFFIX = '_thumb_bare.jpg'
CHIP_THUMB_SUFFIX  = '_chip_thumb.jpg'


VS_EXEMPLARS_KEY = 'vs_exemplars'
INTRA_OCCUR_KEY = 'intra_occurrence'

HARD_NOTE_TAG = '<HARDCASE>'

# HACK
COMPUTER_NAME = ut.get_computer_name()
if COMPUTER_NAME in ['ibeis.cs.uic.edu']:
    _DEFAULT_WILDBOOK_TARGET = 'lewa3'
elif COMPUTER_NAME in ['Leviathan']:
    _DEFAULT_WILDBOOK_TARGET = 'lewa3'
elif COMPUTER_NAME in ['maasai', 'quagga', 'xadmin-Nitro-AN515-51']:
    _DEFAULT_WILDBOOK_TARGET = 'quagga.princeton.edu'
else:
    _DEFAULT_WILDBOOK_TARGET = 'ibeis'
WILDBOOK_TARGET = ut.get_argval('--wildbook-target', type_=str, default=_DEFAULT_WILDBOOK_TARGET,
                                help_='specify the Wildbook target deployment')


class ZIPPED_URLS(object):
    PZ_MTEST       = 'https://lev.cs.rpi.edu/public/databases/PZ_MTEST.zip'
    NAUTS          = 'https://lev.cs.rpi.edu/public/databases/NAUT_test.zip'
    WDS            = 'https://lev.cs.rpi.edu/public/databases/wd_peter2.zip'
    PZ_DISTINCTIVE = 'https://lev.cs.rpi.edu/public/models/distinctivness_zebra_plains.zip'  # DEPRICATE
    GZ_DISTINCTIVE = 'https://lev.cs.rpi.edu/public/models/distinctivness_zebra_grevys.zip'  # DEPRICATE
    DF_CURVRANK    = 'https://lev.cs.rpi.edu/public/databases/testdb_curvrank.zip'


# Turn off features at Lewa :(
SIMPLIFY_INTERFACE = (ut.get_computer_name() == 'ibeis.cs.uic.edu') or ut.get_argflag('--simplify')


# For candidacy document
DBNAME_ALIAS = {
    #'NNP_MasterGIRM_core': 'NNP_GIRM'
    #'NNP_MasterGIRM_core': 'GIRM',
    'NNP_MasterGIRM_core': 'GIRM',
    'PZ_Master1': 'PZ',
    'GZ_Master1': 'GZ',
    'GIRM_Master1': 'GIRM',
    'GZ_ALL': 'GZ',
}


class TEST_SPECIES(object):
    BEAR_POLAR      = 'bear_polar'
    BUILDING        = 'building'
    GIR_RETICULATED = 'giraffe_reticulated'
    GIR_MASAI       = 'giraffe_masai'
    WHALE_FLUKE     = 'whale_fluke',
    WHALE_HUMPBACK  = 'whale_humpback',
    ZEB_GREVY       = 'zebra_grevys'
    ZEB_HYBRID      = 'zebra_hybrid'
    ZEB_PLAIN       = 'zebra_plains'
    LYNX            = 'lynx'
    CHEETAH         = 'cheetah'
    SHARK_SANDTIGER = 'shark_sandtiger'
    UNKNOWN         = UNKNOWN


SPECIES_WITH_DETECTORS = (
    TEST_SPECIES.ZEB_GREVY,
    TEST_SPECIES.ZEB_PLAIN,
    TEST_SPECIES.GIR_RETICULATED,
    TEST_SPECIES.GIR_MASAI,
    TEST_SPECIES.WHALE_FLUKE,
    TEST_SPECIES.WHALE_HUMPBACK,
    TEST_SPECIES.LYNX,
    TEST_SPECIES.CHEETAH,
)

SPECIES_MAPPING = {
    'bear_polar'          :          ('PB', 'Polar Bear'),
    'bird'                :        ('BIRD', 'Bird (Generic)'),
    'bird_generic'        :          (None, 'bird'),
    'building'            :    ('BUILDING', 'Building'),
    'car'                 :         ('CAR', 'Car'),
    'camel'               :       ('CAMEL', 'Camel'),
    'crane'               :       ('CRANE', 'Crane'),
    'cheetah'             :        ('CHTH', 'Cheetah'),
    'dog_wild'            :          ('WD', 'Wild Dog'),
    'dolphin_fin'         :          ('DF', 'Dolphin Fin'),
    'elephant_savanna'    :        ('ELEP', 'Elephant (Savanna)'),
    'elephant_savannah'   :          (None, 'elephant_savanna'),
    'frog'                :        ('FROG', 'Frog'),
    'gazelle'             :         ('GAZ', 'Gazelle (Generic)'),
    'gazelle_generic'     :          (None, 'gazelle'),
    'giraffe_masai'       :        ('GIRM', 'Giraffe (Masai)'),
    'giraffe_massai'      :          (None, 'giraffe_masai'),
    'giraffe_maasai'      :          (None, 'giraffe_masai'),
    'giraffe_reticulated' :         ('GIR', 'Giraffe (Reticulated)'),
    'goat'                :        ('GOAT', 'Goat (Generic)'),
    'goat_generic'        :          (None, 'goat'),
    'hyena'               :       ('HYENA', 'Hyena'),
    'indistinct'          :  ('INDISTINCT', 'INDISTINCT'),
    'jaguar'              :         ('JAG', 'Jaguar'),
    'leopard'             :        ('LOEP', 'Leopard'),
    'lion'                :        ('LION', 'Lion'),
    'lionfish'            :          ('LF', 'Lionfish'),
    'lynx'                :        ('LYNX', 'Lynx'),
    'nautilus'            :        ('NAUT', 'Nautilus'),
    'ostrich'             :        ('OST',  'Ostrich'),
    'other'               :       ('OTHER', 'Other'),
    'person'              :      ('PERSON', 'Person'),
    'rhino_black'         :      ('BRHINO', 'Rhino (Black)'),
    'rhino_white'         :      ('WRHINO', 'Rhino (White)'),
    'seal_saimma_ringed'  :       ('SEAL2', 'Seal (Siamaa Ringed)'),
    'seal_spotted'        :       ('SEAL1', 'Seal (Spotted)'),
    'shark_sandtiger'     :         ('STS', 'Sand Tiger Shark'),
    'snail'               :       ('SNAIL', 'Snail'),
    'snow_leopard'        :       ('SLEOP', 'Snow Leopard'),
    'tiger'               :       ('TIGER', 'Tiger'),
    'turtle_sea'          :          ('ST', 'Sea Turtle (Generic)'),
    'turtle_sea_generic'  :          (None, 'turtle_sea'),
    'turtle_hawksbill'    :        ('STHB', 'Sea Turtle (Hawksbill)'),
    'turtle_green'        :         ('STG', 'Sea Turtle (Green)'),
    'toads_wyoming'       :      ('WYTOAD', 'Toad (Wyoming)'),
    'unspecified'         : ('UNSPECIFIED', 'UNSPECIFIED'),
    'unspecified_animal'  :          (None, 'unspecified'),
    'warthog'             :        ('WART', 'Warthog'),
    'water_buffalo'       :        ('BUFF', 'Water Buffalo'),
    'wildebeest'          :          ('WB', 'Wildebeest'),
    'whale_fluke'         :          ('WF', 'Whale Fluke'),
    'whale_humpback'      :          ('HW', 'Humpback Whale'),
    'whale_shark'         :          ('WS', 'Whale Shark'),
    'zebra_grevys'        :          ('GZ', 'Zebra (Grevy\'s)'),
    'zebra_hybrid'        :          ('HZ', 'Zebra (Hybrid)'),
    'zebra_plains'        :          ('PZ', 'Zebra (Plains)'),
    UNKNOWN               :     ('UNKNOWN', 'UNSPECIFIED'),
}


class _ConstHelper(type):
    """
    Adds code and nice constants to an integer version of a class
    """
    def __new__(cls, name, parents, dct):
        """
        cls = META_DECISION
        code_cls = META_DECISION_CODE
        """
        class CODE(object):
            pass

        class NICE(object):
            pass

        for key in dct.keys():
            if key.isupper():
                value = dct[key]
                if value is None or isinstance(value, int):
                    code = dct['INT_TO_CODE'][value]
                    nice = dct['INT_TO_NICE'][value]
                    setattr(CODE, key, code)
                    setattr(NICE, key, nice)

        dct['CODE'] = CODE
        dct['NICE'] = NICE
        # we need to call type.__new__ to complete the initialization
        return super(_ConstHelper, cls).__new__(cls, name, parents, dct)


@six.add_metaclass(_ConstHelper)
class EVIDENCE_DECISION(object):
    """
    TODO: change to EVIDENCE_DECISION / VISUAL_DECISION
    Enumerated types of review codes and texts

    Notes:
        Unreviewed: Not comparared yet.
        nomatch: Visually comparable and the different
        match: Visually comparable and the same
        notcomp: Not comparable means it is actually impossible to determine.
        unknown: means that it was reviewed, but we just can't figure it out.
    """
    UNREVIEWED   = None
    NEGATIVE     = 0
    POSITIVE     = 1
    INCOMPARABLE = 2
    UNKNOWN      = 3

    INT_TO_CODE = ut.odict([
        (POSITIVE       , 'match'),
        (NEGATIVE       , 'nomatch'),
        (INCOMPARABLE   , 'notcomp'),
        (UNKNOWN        , 'unknown'),
        (UNREVIEWED     , 'unreviewed'),
    ])

    INT_TO_NICE = ut.odict([
        (POSITIVE       , 'Positive'),
        (NEGATIVE       , 'Negative'),
        (INCOMPARABLE   , 'Incomparable'),
        (UNKNOWN        , 'Unknown'),
        (UNREVIEWED     , 'Unreviewed'),
    ])

    CODE_TO_NICE = ut.map_keys(INT_TO_CODE, INT_TO_NICE)
    CODE_TO_INT = ut.invert_dict(INT_TO_CODE)
    NICE_TO_CODE = ut.invert_dict(CODE_TO_NICE)
    NICE_TO_INT  = ut.invert_dict(INT_TO_NICE)

    MATCH_CODE = CODE_TO_INT


@six.add_metaclass(_ConstHelper)
class META_DECISION(object):
    """
    Enumerated types of review codes and texts

    Notes:
        unreviewed: we dont have a meta decision
        same: we know this is the same animal through non-visual means
        diff: we know this is the different animal through non-visual means

    Example:
        >>> # ENABLE_DOCTEST
        >>> from ibeis.constants import *  # NOQA
        >>> assert hasattr(META_DECISION, 'CODE')
        >>> assert hasattr(META_DECISION, 'NICE')
        >>> code1 = META_DECISION.INT_TO_CODE[META_DECISION.NULL]
        >>> code2 = META_DECISION.CODE.NULL
        >>> assert code1 == code2
        >>> nice1 = META_DECISION.INT_TO_NICE[META_DECISION.NULL]
        >>> nice2 = META_DECISION.NICE.NULL
        >>> assert nice1 == nice2
    """
    NULL = None
    DIFF = 0
    SAME = 1
    INT_TO_CODE = ut.odict([
        (NULL , 'null'),
        (DIFF , 'diff'),
        (SAME , 'same'),
    ])
    INT_TO_NICE = ut.odict([
        (NULL , 'NULL'),
        (DIFF , 'Different'),
        (SAME , 'Same'),
    ])
    CODE_TO_NICE = ut.map_keys(INT_TO_CODE, INT_TO_NICE)
    CODE_TO_INT = ut.invert_dict(INT_TO_CODE)
    NICE_TO_CODE = ut.invert_dict(CODE_TO_NICE)
    NICE_TO_INT  = ut.invert_dict(INT_TO_NICE)


@six.add_metaclass(_ConstHelper)
class CONFIDENCE(object):
    UNKNOWN         = None
    GUESSING        = 1
    NOT_SURE        = 2
    PRETTY_SURE     = 3
    ABSOLUTELY_SURE = 4

    INT_TO_CODE = ut.odict([
        (ABSOLUTELY_SURE, 'absolutely_sure'),
        (PRETTY_SURE, 'pretty_sure'),
        (NOT_SURE, 'not_sure'),
        (GUESSING, 'guessing'),
        (UNKNOWN, 'unspecified'),
    ])

    INT_TO_NICE = ut.odict([
        (ABSOLUTELY_SURE, 'Doubtless'),
        (PRETTY_SURE, 'Sure'),
        (NOT_SURE, 'Unsure'),
        (GUESSING, 'Guessing'),
        (UNKNOWN, 'Unspecified'),
    ])

    CODE_TO_NICE = ut.map_keys(INT_TO_CODE, INT_TO_NICE)
    CODE_TO_INT = ut.invert_dict(INT_TO_CODE)
    NICE_TO_CODE = ut.invert_dict(CODE_TO_NICE)
    NICE_TO_INT  = ut.invert_dict(INT_TO_NICE)


@six.add_metaclass(_ConstHelper)
class QUAL(object):
    EXCELLENT = 5
    GOOD      = 4
    OK        = 3
    POOR      = 2
    JUNK      = 1
    UNKNOWN   = None

    INT_TO_CODE = ut.odict([
        (EXCELLENT , 'excellent'),
        (GOOD      , 'good'),
        (OK        , 'ok'),
        (POOR      , 'poor'),
        (JUNK      , 'junk'),
        (UNKNOWN   , 'unspecified'),
    ])

    INT_TO_NICE = ut.odict([
        (EXCELLENT , 'Excellent'),
        (GOOD      , 'Good'),
        (OK        , 'OK'),
        (POOR      , 'Poor'),
        (JUNK      , 'Junk'),
        (UNKNOWN   , 'Unspecified'),
    ])

    CODE_TO_NICE = ut.map_keys(INT_TO_CODE, INT_TO_NICE)
    CODE_TO_INT = ut.invert_dict(INT_TO_CODE)
    NICE_TO_CODE = ut.invert_dict(CODE_TO_NICE)
    NICE_TO_INT  = ut.invert_dict(INT_TO_NICE)


@six.add_metaclass(_ConstHelper)
class VIEW(object):
    """
    categorical viewpoint using the faces of a Rhombicuboctahedron

    References:
        https://en.wikipedia.org/wiki/Rhombicuboctahedron
    """
    UNKNOWN = None
    R  = 1
    FR = 2
    F  = 3
    FL = 4
    L  = 5
    BL = 6
    B  = 7
    BR = 8

    U   = 9
    UF  = 10
    UB  = 11
    UL  = 12
    UR  = 13
    UFL = 14
    UFR = 15
    UBL = 16
    UBR = 17

    D   = 18
    DF  = 19
    DB  = 20
    DL  = 21
    DR  = 22
    DFL = 23
    DFR = 24
    DBL = 25
    DBR = 26

    INT_TO_CODE = ut.odict([
        (UNKNOWN, 'unknown'),
        (R,  'right'),
        (FR, 'frontright'),
        (F,  'front'),
        (FL, 'frontleft'),
        (L,  'left'),
        (BL, 'backleft'),
        (B,  'back'),
        (BR, 'backright'),

        (U,    'up'),
        (UF,   'upfront'),
        (UB,   'upback'),
        (UL,   'upleft'),
        (UR,   'upright'),
        (UFL,  'upfrontleft'),
        (UFR,  'upfrontright'),
        (UBL,  'upbackleft'),
        (UBR,  'upbackright'),

        (D,    'down'),
        (DF,   'downfront'),
        (DB,   'downback'),
        (DL,   'downleft'),
        (DR,   'downright'),
        (DFL,  'downfrontleft'),
        (DFR,  'downfrontright'),
        (DBL,  'downbackleft'),
        (DBR,  'downbackright'),

    ])

    INT_TO_NICE = ut.odict([
        (UNKNOWN, 'Unknown'),
        (R,  'Right'),
        (FR, 'Front-Right'),
        (F,  'Front'),
        (FL, 'Front-Left'),
        (L,  'Left'),
        (BL, 'Back-Left'),
        (B,  'Back'),
        (BR, 'Back-Right'),

        (U,    'Up'),
        (UF,   'Up-Front'),
        (UB,   'Up-Back'),
        (UL,   'Up-Left'),
        (UR,   'Up-Right'),
        (UFL,  'Up-Front-Left'),
        (UFR,  'Up-Front-Right'),
        (UBL,  'Up-Back-Left'),
        (UBR,  'Up-Back-Right'),

        (D,    'Down'),
        (DF,   'Down-Front'),
        (DB,   'Down-Back'),
        (DL,   'Down-Left'),
        (DR,   'Down-Right'),
        (DFL,  'Down-Front-Left'),
        (DFR,  'Down-Front-Right'),
        (DBL,  'Down-Back-Left'),
        (DBR,  'Down-Back-Right'),
    ])

    CODE_TO_NICE = ut.map_keys(INT_TO_CODE, INT_TO_NICE)
    CODE_TO_INT = ut.invert_dict(INT_TO_CODE)
    NICE_TO_CODE = ut.invert_dict(CODE_TO_NICE)
    NICE_TO_INT  = ut.invert_dict(INT_TO_NICE)

    DIST = {
        # DIST 0 PAIRS
        (B, B): 0, (BL, BL): 0, (BR, BR): 0, (D, D): 0, (DB, DB): 0,
        (DBL, DBL): 0, (DBR, DBR): 0, (DF, DF): 0, (DFL, DFL): 0,
        (DFR, DFR): 0, (DL, DL): 0, (DR, DR): 0, (F, F): 0, (FL, FL): 0,
        (FR, FR): 0, (L, L): 0, (R, R): 0, (U, U): 0, (UB, UB): 0,
        (UBL, UBL): 0, (UBR, UBR): 0, (UF, UF): 0, (UFL, UFL): 0,
        (UFR, UFR): 0, (UL, UL): 0, (UR, UR): 0,

        # DIST 1 PAIRS
        (B, BL): 1, (B, BR): 1, (B, DB): 1, (B, DBL): 1, (B, DBR): 1,
        (B, UB): 1, (B, UBL): 1, (B, UBR): 1, (BL, DBL): 1, (BL, L): 1,
        (BL, UBL): 1, (BR, DBR): 1, (BR, R): 1, (BR, UBR): 1, (D, DB): 1,
        (D, DBL): 1, (D, DBR): 1, (D, DF): 1, (D, DFL): 1, (D, DFR): 1,
        (D, DL): 1, (D, DR): 1, (DB, DBL): 1, (DB, DBR): 1, (DBL, DL): 1,
        (DBL, L): 1, (DBR, DR): 1, (DBR, R): 1, (DF, DFL): 1, (DF, DFR): 1,
        (DF, F): 1, (DFL, DL): 1, (DFL, F): 1, (DFL, FL): 1, (DFL, L): 1,
        (DFR, DR): 1, (DFR, F): 1, (DFR, FR): 1, (DFR, R): 1, (DL, L): 1,
        (DR, R): 1, (F, FL): 1, (F, FR): 1, (F, UF): 1, (F, UFL): 1,
        (F, UFR): 1, (FL, L): 1, (FL, UFL): 1, (FR, R): 1, (FR, UFR): 1,
        (L, UBL): 1, (L, UFL): 1, (L, UL): 1, (R, UBR): 1, (R, UFR): 1,
        (R, UR): 1, (U, UB): 1, (U, UBL): 1, (U, UBR): 1, (U, UF): 1,
        (U, UFL): 1, (U, UFR): 1, (U, UL): 1, (U, UR): 1, (UB, UBL): 1,
        (UB, UBR): 1, (UBL, UL): 1, (UBR, UR): 1, (UF, UFL): 1, (UF, UFR): 1,
        (UFL, UL): 1, (UFR, UR): 1,

        # DIST 2 PAIRS
        (B, D): 2, (B, DL): 2, (B, DR): 2, (B, L): 2, (B, R): 2, (B, U): 2,
        (B, UL): 2, (B, UR): 2, (BL, BR): 2, (BL, D): 2, (BL, DB): 2,
        (BL, DBR): 2, (BL, DFL): 2, (BL, DL): 2, (BL, FL): 2, (BL, U): 2,
        (BL, UB): 2, (BL, UBR): 2, (BL, UFL): 2, (BL, UL): 2, (BR, D): 2,
        (BR, DB): 2, (BR, DBL): 2, (BR, DFR): 2, (BR, DR): 2, (BR, FR): 2,
        (BR, U): 2, (BR, UB): 2, (BR, UBL): 2, (BR, UFR): 2, (BR, UR): 2,
        (D, F): 2, (D, FL): 2, (D, FR): 2, (D, L): 2, (D, R): 2, (DB, DF): 2,
        (DB, DFL): 2, (DB, DFR): 2, (DB, DL): 2, (DB, DR): 2, (DB, L): 2,
        (DB, R): 2, (DB, UB): 2, (DB, UBL): 2, (DB, UBR): 2, (DBL, DBR): 2,
        (DBL, DF): 2, (DBL, DFL): 2, (DBL, DFR): 2, (DBL, DR): 2, (DBL, FL): 2,
        (DBL, UB): 2, (DBL, UBL): 2, (DBL, UBR): 2, (DBL, UFL): 2,
        (DBL, UL): 2, (DBR, DF): 2, (DBR, DFL): 2, (DBR, DFR): 2, (DBR, DL): 2,
        (DBR, FR): 2, (DBR, UB): 2, (DBR, UBL): 2, (DBR, UBR): 2,
        (DBR, UFR): 2, (DBR, UR): 2, (DF, DL): 2, (DF, DR): 2, (DF, FL): 2,
        (DF, FR): 2, (DF, L): 2, (DF, R): 2, (DF, UF): 2, (DF, UFL): 2,
        (DF, UFR): 2, (DFL, DFR): 2, (DFL, DR): 2, (DFL, FR): 2, (DFL, UBL): 2,
        (DFL, UF): 2, (DFL, UFL): 2, (DFL, UFR): 2, (DFL, UL): 2, (DFR, DL): 2,
        (DFR, FL): 2, (DFR, UBR): 2, (DFR, UF): 2, (DFR, UFL): 2,
        (DFR, UFR): 2, (DFR, UR): 2, (DL, DR): 2, (DL, F): 2, (DL, FL): 2,
        (DL, UBL): 2, (DL, UFL): 2, (DL, UL): 2, (DR, F): 2, (DR, FR): 2,
        (DR, UBR): 2, (DR, UFR): 2, (DR, UR): 2, (F, L): 2, (F, R): 2,
        (F, U): 2, (F, UL): 2, (F, UR): 2, (FL, FR): 2, (FL, U): 2,
        (FL, UBL): 2, (FL, UF): 2, (FL, UFR): 2, (FL, UL): 2, (FR, U): 2,
        (FR, UBR): 2, (FR, UF): 2, (FR, UFL): 2, (FR, UR): 2, (L, U): 2,
        (L, UB): 2, (L, UF): 2, (R, U): 2, (R, UB): 2, (R, UF): 2, (UB, UF): 2,
        (UB, UFL): 2, (UB, UFR): 2, (UB, UL): 2, (UB, UR): 2, (UBL, UBR): 2,
        (UBL, UF): 2, (UBL, UFL): 2, (UBL, UFR): 2, (UBL, UR): 2, (UBR, UF): 2,
        (UBR, UFL): 2, (UBR, UFR): 2, (UBR, UL): 2, (UF, UL): 2, (UF, UR): 2,
        (UFL, UFR): 2, (UFL, UR): 2, (UFR, UL): 2, (UL, UR): 2,

        # DIST 3 PAIRS
        (B, DF): 3, (B, DFL): 3, (B, DFR): 3, (B, FL): 3, (B, FR): 3,
        (B, UF): 3, (B, UFL): 3, (B, UFR): 3, (BL, DF): 3, (BL, DFR): 3,
        (BL, DR): 3, (BL, F): 3, (BL, R): 3, (BL, UF): 3, (BL, UFR): 3,
        (BL, UR): 3, (BR, DF): 3, (BR, DFL): 3, (BR, DL): 3, (BR, F): 3,
        (BR, L): 3, (BR, UF): 3, (BR, UFL): 3, (BR, UL): 3, (D, UB): 3,
        (D, UBL): 3, (D, UBR): 3, (D, UF): 3, (D, UFL): 3, (D, UFR): 3,
        (D, UL): 3, (D, UR): 3, (DB, F): 3, (DB, FL): 3, (DB, FR): 3, (DB, U): 3,
        (DB, UFL): 3, (DB, UFR): 3, (DB, UL): 3, (DB, UR): 3, (DBL, F): 3,
        (DBL, FR): 3, (DBL, R): 3, (DBL, U): 3, (DBL, UF): 3, (DBL, UR): 3,
        (DBR, F): 3, (DBR, FL): 3, (DBR, L): 3, (DBR, U): 3, (DBR, UF): 3,
        (DBR, UL): 3, (DF, U): 3, (DF, UBL): 3, (DF, UBR): 3, (DF, UL): 3,
        (DF, UR): 3, (DFL, R): 3, (DFL, U): 3, (DFL, UB): 3, (DFL, UR): 3,
        (DFR, L): 3, (DFR, U): 3, (DFR, UB): 3, (DFR, UL): 3, (DL, FR): 3,
        (DL, R): 3, (DL, U): 3, (DL, UB): 3, (DL, UBR): 3, (DL, UF): 3,
        (DL, UFR): 3, (DR, FL): 3, (DR, L): 3, (DR, U): 3, (DR, UB): 3,
        (DR, UBL): 3, (DR, UF): 3, (DR, UFL): 3, (F, UB): 3, (F, UBL): 3,
        (F, UBR): 3, (FL, R): 3, (FL, UB): 3, (FL, UBR): 3, (FL, UR): 3,
        (FR, L): 3, (FR, UB): 3, (FR, UBL): 3, (FR, UL): 3, (L, UBR): 3,
        (L, UFR): 3, (L, UR): 3, (R, UBL): 3, (R, UFL): 3, (R, UL): 3,

        # DIST 4 PAIRS
        (B, F): 4, (BL, FR): 4, (BR, FL): 4, (D, U): 4, (DB, UF): 4,
        (DBL, UFR): 4, (DBR, UFL): 4, (DF, UB): 4, (DFL, UBR): 4,
        (DFR, UBL): 4, (DL, UR): 4, (DR, UL): 4, (L, R): 4,

        # UNDEFINED DIST PAIRS
        (B, UNKNOWN): None, (BL, UNKNOWN): None, (BR, UNKNOWN): None,
        (D, UNKNOWN): None, (DB, UNKNOWN): None, (DBL, UNKNOWN): None,
        (DBR, UNKNOWN): None, (DF, UNKNOWN): None, (DFL, UNKNOWN): None,
        (DFR, UNKNOWN): None, (DL, UNKNOWN): None, (DR, UNKNOWN): None,
        (F, UNKNOWN): None, (FL, UNKNOWN): None, (FR, UNKNOWN): None,
        (L, UNKNOWN): None, (R, UNKNOWN): None, (U, UNKNOWN): None,
        (UB, UNKNOWN): None, (UBL, UNKNOWN): None, (UBR, UNKNOWN): None,
        (UF, UNKNOWN): None, (UFL, UNKNOWN): None, (UFR, UNKNOWN): None,
        (UL, UNKNOWN): None, (UNKNOWN, B): None, (UNKNOWN, BL): None,
        (UNKNOWN, BR): None, (UNKNOWN, D): None, (UNKNOWN, DB): None,
        (UNKNOWN, DBL): None, (UNKNOWN, DBR): None, (UNKNOWN, DF): None,
        (UNKNOWN, DFL): None, (UNKNOWN, DFR): None, (UNKNOWN, DL): None,
        (UNKNOWN, DR): None, (UNKNOWN, F): None, (UNKNOWN, FL): None,
        (UNKNOWN, FR): None, (UNKNOWN, L): None, (UNKNOWN, R): None,
        (UNKNOWN, U): None, (UNKNOWN, UB): None, (UNKNOWN, UBL): None,
        (UNKNOWN, UBR): None, (UNKNOWN, UF): None, (UNKNOWN, UFL): None,
        (UNKNOWN, UFR): None, (UNKNOWN, UL): None, (UNKNOWN, UR): None,
        (UR, UNKNOWN): None, (UNKNOWN, UNKNOWN): None,
    }
    # make distance symmetric
    for (f1, f2), d in list(DIST.items()):
        DIST[(f2, f1)] = d


if __name__ == '__main__':
    r"""
    CommandLine:
        python -m ibeis.constants
        python -m ibeis.constants --allexamples
    """
    import multiprocessing
    multiprocessing.freeze_support()  # for win32
    import utool as ut  # NOQA
    ut.doctest_funcs()



# -*- coding: utf-8 -*-
"""
Autogenerated IBEISController functions

TemplateInfo:
    autogen_time = 13:34:34 2015/04/28
    autogen_key = gar

ToRegenerate:
    python -m ibeis.templates.template_generator --key gar --Tcfg with_web_api=True with_api_cache=False with_deleters=True no_extern_deleters=True --diff
    python -m ibeis.templates.template_generator --key gar --Tcfg with_web_api=True with_api_cache=False with_deleters=True no_extern_deleters=True --write
"""
from __future__ import absolute_import, division, print_function
from six.moves import zip
from ibeis import constants as const
import utool as ut
from ibeis.control import controller_inject
from ibeis.control import accessor_decors
print, rrr, profile = ut.inject2(__name__)

# Create dectorator to inject functions in this module into the IBEISController
CLASS_INJECT_KEY, register_ibs_method = controller_inject.make_ibs_register_decorator(__name__)


register_api   = controller_inject.get_ibeis_flask_api(__name__)


def testdata_ibs(defaultdb='testdb1'):
    import ibeis
    ibs = ibeis.opendb(defaultdb=defaultdb)
    config2_ = None  # qreq_.qparams
    return ibs, config2_

# AUTOGENED CONSTANTS:
GAR_ROWID        = 'gar_rowid'
ANNOTGROUP_ROWID = 'annotgroup_rowid'
ANNOT_ROWID      = 'annot_rowid'


@register_ibs_method
# @register_api('/api/gar/', methods=['GET'])
def _get_all_gar_rowids(ibs):
    """ all_gar_rowids <- gar.get_all_rowids()

    Returns:
        list_ (list): unfiltered gar_rowids

    TemplateInfo:
        Tider_all_rowids
        tbl = gar

    Example:
        >>> # ENABLE_DOCTEST
        >>> from ibeis.control.manual_garelate_funcs import *  # NOQA
        >>> ibs, config2_ = testdata_ibs()
        >>> ibs._get_all_gar_rowids()
    """
    all_gar_rowids = ibs.db.get_all_rowids(
        const.GA_RELATION_TABLE)
    return all_gar_rowids


@register_ibs_method
# @register_api('/api/gar/', methods=['POST'])
def add_gar(ibs, annotgroup_rowid_list, aid_list):
    """
    Returns:
        returns gar_rowid_list of added (or already existing gars)

    TemplateInfo:
        Tadder_native
        tbl = gar
    """
    # WORK IN PROGRESS
    colnames = (ANNOTGROUP_ROWID, ANNOT_ROWID,)
    #if aid_list is None:
    #    aid_list = [None] * len(annotgroup_rowid_list)
    params_iter = (
        (annotgroup_rowid, aid,)
        for (annotgroup_rowid, aid,) in
        zip(annotgroup_rowid_list, aid_list)
    )
    get_rowid_from_superkey = ibs.get_gar_rowid_from_superkey
    # FIXME: encode superkey paramx
    superkey_paramx = (0, 1)
    gar_rowid_list = ibs.db.add_cleanly(
        const.GA_RELATION_TABLE, colnames, params_iter, get_rowid_from_superkey, superkey_paramx)
    return gar_rowid_list


@register_ibs_method
# @register_api('/api/gar/', methods=['DELETE'])
def delete_gar(ibs, gar_rowid_list, config2_=None):
    """ gar.delete(gar_rowid_list)

    delete gar rows

    Args:
        gar_rowid_list

    Returns:
        int: num_deleted

    TemplateInfo:
        Tdeleter_native_tbl
        tbl = gar

    Example:
        >>> # DISABLE_DOCTEST
        >>> from ibeis.control.manual_garelate_funcs import *  # NOQA
        >>> ibs, config2_ = testdata_ibs()
        >>> gar_rowid_list = ibs._get_all_gar_rowids()[:2]
        >>> num_deleted = ibs.delete_gar(gar_rowid_list)
        >>> print('num_deleted = %r' % (num_deleted,))
    """
    #from ibeis.algo.preproc import preproc_gar
    # NO EXTERN IMPORT
    if ut.VERBOSE:
        print('[ibs] deleting %d gar rows' %
              len(gar_rowid_list))
    # Prepare: Delete externally stored data (if any)
    #preproc_gar.on_delete(ibs, gar_rowid_list, config2_=config2_)
    # NO EXTERN DELETE
    # Finalize: Delete self
    ibs.db.delete_rowids(
        const.GA_RELATION_TABLE, gar_rowid_list)
    num_deleted = len(
        ut.filter_Nones(gar_rowid_list))
    return num_deleted


@register_ibs_method
@accessor_decors.getter_1to1
# @register_api('/api/gar/annot/rowid/', methods=['GET'])
def get_gar_aid(ibs, gar_rowid_list, eager=True, nInput=None):
    """ aid_list <- gar.aid[gar_rowid_list]

    gets data from the "native" column "aid" in the "gar" table

    Args:
        gar_rowid_list (list):

    Returns:
        list: aid_list

    TemplateInfo:
        Tgetter_table_column
        col = aid
        tbl = gar

    Example:
        >>> # ENABLE_DOCTEST
        >>> from ibeis.control.manual_garelate_funcs import *  # NOQA
        >>> ibs, config2_ = testdata_ibs()
        >>> gar_rowid_list = ibs._get_all_gar_rowids()
        >>> eager = True
        >>> aid_list = ibs.get_gar_aid(gar_rowid_list, eager=eager)
        >>> assert len(gar_rowid_list) == len(aid_list)
    """
    id_iter = gar_rowid_list
    colnames = (ANNOT_ROWID,)
    aid_list = ibs.db.get(const.GA_RELATION_TABLE, colnames,
                          id_iter, id_colname='rowid', eager=eager, nInput=nInput)
    return aid_list


@register_ibs_method
@accessor_decors.getter_1to1
# @register_api('/api/gar/annotgroup/rowid/', methods=['GET'])
def get_gar_annotgroup_rowid(ibs, gar_rowid_list, eager=True, nInput=None):
    """ annotgroup_rowid_list <- gar.annotgroup_rowid[gar_rowid_list]

    gets data from the "native" column "annotgroup_rowid" in the "gar" table

    Args:
        gar_rowid_list (list):

    Returns:
        list: annotgroup_rowid_list

    TemplateInfo:
        Tgetter_table_column
        col = annotgroup_rowid
        tbl = gar

    Example:
        >>> # ENABLE_DOCTEST
        >>> from ibeis.control.manual_garelate_funcs import *  # NOQA
        >>> ibs, config2_ = testdata_ibs()
        >>> gar_rowid_list = ibs._get_all_gar_rowids()
        >>> eager = True
        >>> annotgroup_rowid_list = ibs.get_gar_annotgroup_rowid(gar_rowid_list, eager=eager)
        >>> assert len(gar_rowid_list) == len(annotgroup_rowid_list)
    """
    id_iter = gar_rowid_list
    colnames = (ANNOTGROUP_ROWID,)
    annotgroup_rowid_list = ibs.db.get(
        const.GA_RELATION_TABLE, colnames, id_iter, id_colname='rowid', eager=eager, nInput=nInput)
    return annotgroup_rowid_list


@register_ibs_method
def get_gar_rowid_from_superkey(ibs, annotgroup_rowid_list, aid_list, eager=True, nInput=None):
    """ gar_rowid_list <- gar[annotgroup_rowid_list, aid_list]

    Args:
        superkey lists: annotgroup_rowid_list, aid_list

    Returns:
        gar_rowid_list

    TemplateInfo:
        Tgetter_native_rowid_from_superkey
        tbl = gar
    """
    colnames = (GAR_ROWID,)
    # FIXME: col_rowid is not correct
    params_iter = zip(annotgroup_rowid_list, aid_list)
    andwhere_colnames = [ANNOTGROUP_ROWID, ANNOT_ROWID]
    gar_rowid_list = ibs.db.get_where_eq(
        const.GA_RELATION_TABLE, colnames, params_iter, andwhere_colnames, eager=eager, nInput=nInput)
    return gar_rowid_list


if __name__ == '__main__':
    """
    CommandLine:
        python -m ibeis.control.manual_garelate_funcs
        python -m ibeis.control.manual_garelate_funcs --allexamples
    """
    import multiprocessing
    multiprocessing.freeze_support()
    import utool as ut
    ut.doctest_funcs()



# -*- coding: utf-8 -*-
"""Interface to Lightnet object proposals."""
from __future__ import absolute_import, division, print_function
import utool as ut
from six.moves import zip
import numpy as np
from os.path import abspath, dirname, expanduser, join, exists  # NOQA
import cv2
from tqdm import tqdm
(print, rrr, profile) = ut.inject2(__name__, '[lightnet]')


if not ut.get_argflag('--no-lightnet'):
    try:
        import torch
        from torchvision import transforms as tf
        import lightnet as ln
    except ImportError:
        print('WARNING Failed to import lightnet. '
              'PyDarknet YOLO detection is unavailable')
        if ut.SUPER_STRICT:
            raise


VERBOSE_LN = ut.get_argflag('--verbln') or ut.VERBOSE


WEIGHT_URL_DICT = {
    'seaturtle'     : 'https://lev.cs.rpi.edu/public/models/detect.lightnet.sea_turtle.weights',
    'hammerhead'    : 'https://lev.cs.rpi.edu/public/models/detect.lightnet.shark_hammerhead.weights',
    'ggr2'          : 'https://lev.cs.rpi.edu/public/models/detect.lightnet.ggr2.weights',
    'lynx'          : 'https://lev.cs.rpi.edu/public/models/detect.lightnet.lynx.weights',

    None            : 'https://lev.cs.rpi.edu/public/models/detect.lightnet.sea_turtle.weights',
}


def _parse_classes_from_weights(url):
    return url.replace('.weights', '.classes')


def _parse_class_list(classes_filepath):
    # Load classes from file into the class list
    assert exists(classes_filepath)
    class_list = []
    with open(classes_filepath) as classes:
        for line in classes.readlines():
            line = line.strip()
            if len(line) > 0:
                class_list.append(line)
    return class_list


def detect_gid_list(ibs, gid_list, verbose=VERBOSE_LN, **kwargs):
    """Detect gid_list with lightnet.

    Args:
        gid_list (list of int): the list of IBEIS image_rowids that need detection

    Kwargs (optional): refer to the Lightnet documentation for configuration settings

    Args:
        ibs (ibeis.IBEISController):  image analysis api
        gid_list (list of int): the list of IBEIS image_rowids that need detection

    Kwargs:
        detector, config_filepath, weights_filepath, verbose

    Yields:
        tuple: (gid, gpath, result_list)
    """
    # Get new gpaths if downsampling
    gpath_list = ibs.get_image_paths(gid_list)

    # Run detection
    results_iter = detect(gpath_list, verbose=verbose, **kwargs)
    # Upscale the results
    _iter = zip(gid_list, results_iter)
    for gid, (gpath, result_list) in _iter:
        # Upscale the results back up to the original image size
        for result in result_list:
            bbox = (result['xtl'], result['ytl'], result['width'], result['height'], )
            bbox_list = [ bbox ]
            bbox = bbox_list[0]
            result['xtl'], result['ytl'], result['width'], result['height'] = bbox
        yield (gid, gpath, result_list)


def _create_network(weight_filepath, class_list, conf_thresh, nms_thresh, network_size):
    """Create the lightnet network."""
    net = ln.models.Yolo(len(class_list), weight_filepath, conf_thresh, nms_thresh)
    net.postprocess.append(ln.data.transform.TensorToBrambox(network_size, class_list))

    net.eval()
    if torch.cuda.is_available():
        net.cuda()

    return net


def _detect(net, img_path, network_size):
    """Perform a detection."""
    # Load image
    img = cv2.imread(img_path)
    im_h, im_w = img.shape[:2]

    img_tf = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
    img_tf = ln.data.transform.Letterbox.apply(img_tf, dimension=network_size)
    img_tf = tf.ToTensor()(img_tf)
    img_tf.unsqueeze_(0)

    if torch.cuda.is_available():
        img_tf = img_tf.cuda()

    # Run detector
    if torch.__version__.startswith('0.3'):
        img_tf = torch.autograd.Variable(img_tf, volatile=True)
        out = net(img_tf)
    else:
        with torch.no_grad():
            out = net(img_tf)
    out = ln.data.transform.ReverseLetterbox.apply(out, network_size, (im_w, im_h))

    return img, out


def detect(gpath_list, config_filepath, weight_filepath, class_filepath, sensitivity,
           verbose=VERBOSE_LN, **kwargs):
    """Detect image filepaths with lightnet.

    Args:
        gpath_list (list of str): the list of image paths that need proposal candidates

    Kwargs (optional): refer to the Lightnet documentation for configuration settings

    Returns:
        iter
    """
    assert config_filepath is None, 'lightnet does not have a config file'

    # Get correct weight if specified with shorthand
    weight_url = None
    if weight_filepath in WEIGHT_URL_DICT:
        weight_url = WEIGHT_URL_DICT[weight_filepath]
        weight_filepath = ut.grab_file_url(weight_url, appname='ibeis',
                                           check_hash=True)

    if class_filepath in WEIGHT_URL_DICT:
        if class_filepath is None and weight_url is not None:
            weight_url_ = weight_url
        else:
            weight_url_ = WEIGHT_URL_DICT[weight_filepath]
        class_url = _parse_classes_from_weights(weight_url_)
        class_filepath = ut.grab_file_url(class_url, appname='ibeis',
                                          check_hash=True, verbose=verbose)

    assert exists(weight_filepath)
    weight_filepath = ut.truepath(weight_filepath)
    assert exists(class_filepath)
    class_filepath = ut.truepath(class_filepath)

    class_list = _parse_class_list(class_filepath)

    network_size = (416, 416)
    conf_thresh = sensitivity
    nms_thresh = 1.0  # Turn off NMS
    network = _create_network(weight_filepath, class_list, conf_thresh,
                              nms_thresh, network_size)

    # Execute detector for each image
    results_list_ = []
    for gpath in tqdm(gpath_list):
        image, output_list = _detect(network, gpath, network_size)
        output_list = output_list[0]

        result_list_ = []
        for output in list(output_list):
            xtl = int(np.around(float(output.x_top_left)))
            ytl = int(np.around(float(output.y_top_left)))
            xbr = int(np.around(float(output.x_top_left + output.width)))
            ybr = int(np.around(float(output.y_top_left + output.height)))
            class_ = output.class_label
            conf = float(output.confidence)
            result_dict = {
                'xtl'        : xtl,
                'ytl'        : ytl,
                'width'      : xbr - xtl,
                'height'     : ybr - ytl,
                'class'      : class_,
                'confidence' : conf,
            }
            result_list_.append(result_dict)
        results_list_.append(result_list_)

    if len(results_list_) != len(gpath_list):
        raise ValueError('Lightnet did not return valid data')

    results_list = zip(gpath_list, results_list_)
    return results_list



import utool as ut
from ibeis.algo.graph import demo
from ibeis.algo.graph.state import (POSTV, NEGTV, INCMP, UNREV)


def test_incomp_inference():
    infr = demo.demodata_infr(num_pccs=0)
    # Make 2 consistent and 2 inconsistent CCs
    infr.add_feedback(( 1,  2), POSTV)
    infr.add_feedback(( 2,  3), POSTV)
    infr.add_feedback(( 3,  4), POSTV)
    infr.add_feedback(( 4,  1), POSTV)
    # -----
    infr.add_feedback((11, 12), POSTV)
    infr.add_feedback((12, 13), POSTV)
    infr.add_feedback((13, 14), POSTV)
    infr.add_feedback((14, 11), POSTV)
    infr.add_feedback((12, 14), NEGTV)
    # -----
    infr.add_feedback((21, 22), POSTV)
    infr.add_feedback((22, 23), POSTV)
    infr.add_feedback((23, 21), NEGTV)
    # -----
    infr.add_feedback((31, 32), POSTV)
    infr.add_feedback((32, 33), POSTV)
    infr.add_feedback((33, 31), POSTV)
    infr.add_feedback(( 2, 32), NEGTV)
    infr.add_feedback(( 3, 33), NEGTV)
    infr.add_feedback((12, 21), NEGTV)
    # -----
    # Incomparable within CCs
    print('==========================')
    infr.add_feedback(( 1, 3), INCMP)
    infr.add_feedback(( 1, 4), INCMP)
    infr.add_feedback(( 1, 2), INCMP)
    infr.add_feedback((11, 13), INCMP)
    infr.add_feedback((11, 14), INCMP)
    infr.add_feedback((11, 12), INCMP)
    infr.add_feedback(( 1, 31), INCMP)
    infr.add_feedback(( 2, 32), INCMP)
    infr.add_feedback((12, 21), INCMP)
    infr.add_feedback((23, 21), INCMP)
    infr.add_feedback((12, 14), INCMP)
    print('Final state:')
    print(ut.repr4(sorted(infr.gen_edge_attrs('decision'))))


def test_unrev_inference():
    infr = demo.demodata_infr(num_pccs=0)
    # Make 2 consistent and 2 inconsistent CCs
    infr.add_feedback(( 1,  2), POSTV)
    infr.add_feedback(( 2,  3), POSTV)
    infr.add_feedback(( 3,  4), POSTV)
    infr.add_feedback(( 4,  1), POSTV)
    # -----
    infr.add_feedback((11, 12), POSTV)
    infr.add_feedback((12, 13), POSTV)
    infr.add_feedback((13, 14), POSTV)
    infr.add_feedback((14, 11), POSTV)
    infr.add_feedback((12, 14), NEGTV)
    # -----
    infr.add_feedback((21, 22), POSTV)
    infr.add_feedback((22, 23), POSTV)
    infr.add_feedback((23, 21), NEGTV)
    # -----
    infr.add_feedback((31, 32), POSTV)
    infr.add_feedback((32, 33), POSTV)
    infr.add_feedback((33, 31), POSTV)
    infr.add_feedback(( 2, 32), NEGTV)
    infr.add_feedback(( 3, 33), NEGTV)
    infr.add_feedback((12, 21), NEGTV)
    # -----
    # Incomparable within CCs
    print('==========================')
    infr.add_feedback(( 1, 3), UNREV)
    infr.add_feedback(( 1, 4), UNREV)
    infr.add_feedback(( 1, 2), UNREV)
    infr.add_feedback((11, 13), UNREV)
    infr.add_feedback((11, 14), UNREV)
    infr.add_feedback((11, 12), UNREV)
    infr.add_feedback(( 1, 31), UNREV)
    infr.add_feedback(( 2, 32), UNREV)
    infr.add_feedback((12, 21), UNREV)
    infr.add_feedback((23, 21), UNREV)
    infr.add_feedback((12, 14), UNREV)
    print('Final state:')
    print(ut.repr4(sorted(infr.gen_edge_attrs('decision'))))


def test_pos_neg():
    infr = demo.demodata_infr(num_pccs=0)
    # Make 3 inconsistent CCs
    infr.add_feedback(( 1,  2), POSTV)
    infr.add_feedback(( 2,  3), POSTV)
    infr.add_feedback(( 3,  4), POSTV)
    infr.add_feedback(( 4,  1), POSTV)
    infr.add_feedback(( 1,  3), NEGTV)
    # -----
    infr.add_feedback((11, 12), POSTV)
    infr.add_feedback((12, 13), POSTV)
    infr.add_feedback((13, 11), NEGTV)
    # -----
    infr.add_feedback((21, 22), POSTV)
    infr.add_feedback((22, 23), POSTV)
    infr.add_feedback((23, 21), NEGTV)
    # -----
    # Fix inconsistency
    infr.add_feedback((23, 21), POSTV)
    # Merge inconsistent CCS
    infr.add_feedback(( 1, 11), POSTV)
    # Negative edge within an inconsistent CC
    infr.add_feedback(( 2, 13), NEGTV)
    # Negative edge external to an inconsistent CC
    infr.add_feedback((12, 21), NEGTV)
    # -----
    # Make inconsistency from positive
    infr.add_feedback((31, 32), POSTV)
    infr.add_feedback((33, 34), POSTV)
    infr.add_feedback((31, 33), NEGTV)
    infr.add_feedback((32, 34), NEGTV)
    infr.add_feedback((31, 34), POSTV)
    # Fix everything
    infr.add_feedback(( 1,  3), POSTV)
    infr.add_feedback(( 2,  4), POSTV)
    infr.add_feedback((32, 34), POSTV)
    infr.add_feedback((31, 33), POSTV)
    infr.add_feedback((13, 11), POSTV)
    infr.add_feedback((23, 21), POSTV)
    infr.add_feedback(( 1, 11), NEGTV)
    print('Final state:')
    print(ut.repr4(sorted(infr.gen_edge_attrs('decision'))))


if __name__ == '__main__':
    r"""
    CommandLine:
        export PYTHONPATH=$PYTHONPATH:/home/joncrall/code/ibeis/ibeis/algo/graph/tests
        python ~/code/ibeis/ibeis/algo/graph/tests/test_graph_iden.py test_pos_neg
        python ~/code/ibeis/ibeis/algo/graph/tests/test_graph_iden.py test_unrev_inference
        python ~/code/ibeis/ibeis/algo/graph/tests/test_graph_iden.py test_incomp_inference
        python ~/code/ibeis/ibeis/algo/graph/tests/test_graph_iden.py --allexamples
    """
    import multiprocessing
    multiprocessing.freeze_support()  # for win32
    import utool as ut  # NOQA
    ut.doctest_funcs()
    # import ubelt as ub



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function
import utool as ut
ut.noinject(__name__, '[guiexcept]')


class NeedsUserInput(Exception):
    def __init__(self, *args):
        super(Exception, self).__init__(*args)


class UserCancel(Exception):
    def __init__(self, *args):
        super(Exception, self).__init__(*args)


class InvalidRequest(Exception):
    def __init__(self, *args):
        super(Exception, self).__init__(*args)



# -*- coding: utf-8 -*-
"""
Autogen:
    python -c "import utool as ut; ut.write_modscript_alias('Tgen.sh', 'ibeis.templates.template_generator')"  # NOQA
    sh Tgen.sh --key part --invert --Tcfg with_getters=True with_setters=True --modfname manual_part_funcs --funcname-filter=age_m  # NOQA
    sh Tgen.sh --key part --invert --Tcfg with_getters=True with_setters=True --modfname manual_part_funcs --funcname-filter=is_  # NOQA
    sh Tgen.sh --key part --invert --Tcfg with_getters=True with_setters=True --modfname manual_part_funcs --funcname-filter=is_ --diff  # NOQA
"""
from __future__ import absolute_import, division, print_function, unicode_literals
import six
import uuid
import numpy as np
from ibeis import constants as const
from ibeis.control import accessor_decors, controller_inject
import utool as ut
from ibeis.control.controller_inject import make_ibs_register_decorator
from ibeis.web import routes_ajax
print, rrr, profile = ut.inject2(__name__)


CLASS_INJECT_KEY, register_ibs_method = make_ibs_register_decorator(__name__)


register_api   = controller_inject.get_ibeis_flask_api(__name__)


PART_NOTE               = 'part_note'
PART_NUM_VERTS          = 'part_num_verts'
PART_ROWID              = 'part_rowid'
PART_TAG_TEXT           = 'part_tag_text'
PART_THETA              = 'part_theta'
PART_VERTS              = 'part_verts'
PART_UUID               = 'part_uuid'
PART_QUALITY            = 'part_quality'
PART_STAGED_FLAG        = 'part_staged_flag'
PART_STAGED_USER_ID     = 'part_staged_user_identity'


# ==========
# IDERS
# ==========


# TODO CACHE THIS AND FIND WHAT IT SHOULD INVALIDATE IT
# ADD PARTS, DELETE PARTS ANYTHING ELSE?
@register_ibs_method
@accessor_decors.ider
def _get_all_part_rowids(ibs):
    r"""
    Returns:
        list_ (list):  all unfiltered part_rowids (part rowids)
    """
    all_part_rowids = ibs.db.get_all_rowids(const.PART_TABLE)
    if all_part_rowids is None:
        all_part_rowids = []
    return all_part_rowids


@register_ibs_method
@accessor_decors.ider
@register_api('/api/part/', methods=['GET'])
def get_valid_part_rowids(ibs, include_only_aid_list=None,
                          is_staged=False, viewpoint='no-filter', minqual=None):
    part_rowid_list = ibs._get_all_part_rowids()

    part_rowid_list = ibs.filter_part_set(
        part_rowid_list, include_only_aid_list=include_only_aid_list,
        is_staged=is_staged, viewpoint=viewpoint, minqual=minqual)
    return part_rowid_list


@register_ibs_method
def get_num_parts(ibs, **kwargs):
    r"""
    Number of valid parts
    """
    part_rowid_list = ibs.get_valid_part_rowids(**kwargs)
    return len(part_rowid_list)


@register_ibs_method
@register_api('/api/part/<rowid>/', methods=['GET'])
def part_src_api(rowid=None):
    r"""
    Returns the base64 encoded image of part <rowid>

    RESTful:
        Method: GET
        URL:    /api/part/<rowid>/
    """
    return routes_ajax.part_src(rowid)


@register_ibs_method
def filter_part_set(ibs, part_rowid_list, include_only_aid_list=None,
                    is_staged=False, viewpoint='no-filter', minqual=None):
    # -- valid part_rowid filtering --

    # filter by is_staged
    if is_staged is True:
        # corresponding unoptimized hack for is_staged
        flag_list = ibs.get_part_staged_flags(part_rowid_list)
        part_rowid_list  = ut.compress(part_rowid_list, flag_list)
    elif is_staged is False:
        flag_list = ibs.get_part_staged_flags(part_rowid_list)
        part_rowid_list  = ut.filterfalse_items(part_rowid_list, flag_list)

    if include_only_aid_list is not None:
        gid_list     = ibs.get_part_gids(part_rowid_list)
        is_valid_gid = [gid in include_only_aid_list for gid in gid_list]
        part_rowid_list     = ut.compress(part_rowid_list, is_valid_gid)
    if viewpoint != 'no-filter':
        viewpoint_list     = ibs.get_part_viewpoints(part_rowid_list)
        is_valid_viewpoint = [viewpoint == flag for flag in viewpoint_list]
        part_rowid_list     = ut.compress(part_rowid_list, is_valid_viewpoint)
    if minqual is not None:
        part_rowid_list = ibs.filter_part_rowids_to_quality(part_rowid_list, minqual, unknown_ok=True)
    part_rowid_list = sorted(part_rowid_list)
    return part_rowid_list


# ==========
# ADDERS
# ==========


@register_ibs_method
@accessor_decors.adder
@register_api('/api/part/', methods=['POST'])
def add_parts(ibs, aid_list, bbox_list=None, theta_list=None,
                detect_confidence_list=None, notes_list=None,
                vert_list=None, part_uuid_list=None, viewpoint_list=None,
                quality_list=None, type_list=None, staged_uuid_list=None,
                staged_user_id_list=None, **kwargs):
    r"""
    Adds an part to annotations

    Args:
        aid_list                 (list): annotation rowids to add part to
        bbox_list                (list): of [x, y, w, h] bounding boxes for each annotation (supply verts instead)
        theta_list               (list): orientations of parts
        vert_list                (list): alternative to bounding box

    Returns:
        list: part_rowid_list

    Ignore:
       detect_confidence_list = None
       notes_list = None
       part_uuid_list = None
       viewpoint_list = None
       quality_list = None
       type_list = None

    RESTful:
        Method: POST
        URL:    /api/part/
    """
    #ut.embed()
    from vtool_ibeis import geometry
    if ut.VERBOSE:
        print('[ibs] adding parts')
    # Prepare the SQL input
    # For import only, we can specify both by setting import_override to True
    assert bool(bbox_list is None) != bool(vert_list is None), (
        'must specify exactly one of bbox_list or vert_list')
    ut.assert_all_not_None(aid_list, 'aid_list')

    if vert_list is None:
        vert_list = geometry.verts_list_from_bboxes_list(bbox_list)
    elif bbox_list is None:
        bbox_list = geometry.bboxes_from_vert_list(vert_list)

    if theta_list is None:
        theta_list = [0.0 for _ in range(len(aid_list))]

    len_bbox    = len(bbox_list)
    len_vert    = len(vert_list)
    len_aid     = len(aid_list)
    len_theta   = len(theta_list)
    try:
        assert len_vert  == len_bbox, 'bbox and verts are not of same size'
        assert len_aid   == len_bbox, 'bbox and aid are not of same size'
        assert len_aid   == len_theta, 'bbox and aid are not of same size'
    except AssertionError as ex:
        ut.printex(ex, key_list=['len_vert', 'len_aid', 'len_bbox'
                                    'len_theta'])
        raise

    if len(aid_list) == 0:
        # nothing is being added
        print('[ibs] WARNING: 0 parts are being added!')
        print(ut.repr2(locals()))
        return []

    if detect_confidence_list is None:
        detect_confidence_list = [0.0 for _ in range(len(aid_list))]
    if notes_list is None:
        notes_list = ['' for _ in range(len(aid_list))]
    if viewpoint_list is None:
        viewpoint_list = [-1.0] * len(aid_list)
    if type_list is None:
        type_list = [const.UNKNOWN] * len(aid_list)

    nVert_list = [len(verts) for verts in vert_list]
    vertstr_list = [six.text_type(verts) for verts in vert_list]
    xtl_list, ytl_list, width_list, height_list = list(zip(*bbox_list))
    assert len(nVert_list) == len(vertstr_list)

    # Build ~~deterministic?~~ random and unique PART ids
    if part_uuid_list is None:
        part_uuid_list = [uuid.uuid4() for _ in range(len(aid_list))]

    if staged_uuid_list is None:
        staged_uuid_list = [None] * len(aid_list)
    is_staged_list = [
        staged_uuid is not None
        for staged_uuid in staged_uuid_list
    ]
    if staged_user_id_list is None:
        staged_user_id_list = [None] * len(aid_list)

    # Define arguments to insert
    colnames = ('part_uuid', 'annot_rowid', 'part_xtl', 'part_ytl',
                'part_width', 'part_height', 'part_theta', 'part_num_verts',
                'part_verts', 'part_viewpoint', 'part_detect_confidence',
                'part_note', 'part_type', 'part_staged_flag', 'part_staged_uuid',
                'part_staged_user_identity')

    check_uuid_flags = [not isinstance(auuid, uuid.UUID) for auuid in part_uuid_list]
    if any(check_uuid_flags):
        pos = ut.list_where(check_uuid_flags)
        raise ValueError('positions %r have malformated UUIDS' % (pos,))

    params_iter = list(zip(part_uuid_list, aid_list, xtl_list, ytl_list,
                            width_list, height_list, theta_list, nVert_list,
                            vertstr_list, viewpoint_list, detect_confidence_list,
                            notes_list, type_list, is_staged_list, staged_uuid_list,
                            staged_user_id_list))

    # Execute add PARTs SQL
    superkey_paramx = (0,)
    get_rowid_from_superkey = ibs.get_part_rowids_from_uuid
    part_rowid_list = ibs.db.add_cleanly(const.PART_TABLE, colnames, params_iter,
                                         get_rowid_from_superkey, superkey_paramx)
    return part_rowid_list


@register_ibs_method
# @register_api('/api/part/rows/', methods=['GET'])
def get_part_rows(ibs, part_rowid_list):
    r"""
    Auto-docstr for 'get_part_rows'
    """
    colnames = ('part_uuid', 'annot_rowid', 'part_xtl', 'part_ytl',
                'part_width', 'part_height', 'part_theta', 'part_num_verts',
                'part_verts', 'part_viewpoint', 'part_detect_confidence',
                'part_note', 'part_quality', 'part_type')
    rows_list = ibs.db.get(const.PART_TABLE, colnames, part_rowid_list,
                           unpack_scalars=False)
    return rows_list


# ==========
# DELETERS
# ==========


@register_ibs_method
@accessor_decors.deleter
@accessor_decors.cache_invalidator(const.PART_TABLE, rowidx=0)
@register_api('/api/part/', methods=['DELETE'])
def delete_parts(ibs, part_rowid_list):
    r"""
    deletes parts from the database

    RESTful:
        Method: DELETE
        URL:    /api/part/

    Args:
        ibs (IBEISController):  ibeis controller object
        part_rowid_list (int):  list of part ids
    """
    if ut.VERBOSE:
        print('[ibs] deleting %d parts' % len(part_rowid_list))
    # delete parent rowid column if exists in part table
    return ibs.db.delete_rowids(const.PART_TABLE, part_rowid_list)


# ==========
# GETTERS
# ==========


@register_ibs_method
@accessor_decors.getter_1to1
@register_api('/api/part/rowid/uuid/', methods=['GET'])
def get_part_rowids_from_uuid(ibs, uuid_list):
    r"""
    Returns:
        list_ (list): part rowids

    RESTful:
        Method: GET
        URL:    /api/part/rowid/uuid/
    """
    # FIXME: MAKE SQL-METHOD FOR NON-ROWID GETTERS
    part_rowids_list = ibs.db.get(const.PART_TABLE, (PART_ROWID,), uuid_list,
                                  id_colname=PART_UUID)
    return part_rowids_list


@register_ibs_method
@accessor_decors.getter_1to1
# @register_api('/api/part/uuid/missing/', methods=['GET'])
def get_part_missing_uuid(ibs, uuid_list):
    r"""
    Returns:
        list_ (list): a list of missing part uuids
    """
    part_rowid_list = ibs.get_part_rowids_from_uuid(uuid_list)
    zipped = zip(part_rowid_list, uuid_list)
    missing_uuid_list = [ uuid for part_rowid, uuid in zipped if part_rowid is None ]
    return missing_uuid_list


@register_ibs_method
@ut.accepts_numpy
@accessor_decors.getter_1toM
@register_api('/api/part/bbox/', methods=['GET'])
def get_part_bboxes(ibs, part_rowid_list):
    r"""
    Returns:
        bbox_list (list):  part bounding boxes in image space

    RESTful:
        Method: GET
        URL:    /api/part/bbox/
    """
    colnames = ('part_xtl', 'part_ytl', 'part_width', 'part_height',)
    bbox_list = ibs.db.get(const.PART_TABLE, colnames, part_rowid_list)
    return bbox_list


@register_ibs_method
@accessor_decors.getter_1to1
@register_api('/api/part/detect/confidence/', methods=['GET'])
def get_part_detect_confidence(ibs, part_rowid_list):
    r"""
    Returns:
        list_ (list): a list confidences that the parts is a valid detection

    RESTful:
        Method: GET
        URL:    /api/part/detect/confidence/
    """
    part_detect_confidence_list = ibs.db.get(const.PART_TABLE,
                                              ('part_detect_confidence',),
                                              part_rowid_list)
    return part_detect_confidence_list


@register_ibs_method
@ut.accepts_numpy
@accessor_decors.getter_1to1
@register_api('/api/part/annot/rowid/', methods=['GET'])
def get_part_aids(ibs, part_rowid_list, assume_unique=False):
    r"""
    Get parent annotation rowids of parts

    Args:
        part_rowid_list (list):

    Returns:
        aid_list (list):  annot rowids

    RESTful:
        Method: GET
        URL:    /api/part/annot/rowid/

    Example:
        >>> # ENABLE_DOCTEST
        >>> from ibeis.control.manual_part_funcs import *  # NOQA
        >>> import ibeis
        >>> ibs = ibeis.opendb('testdb1')
        >>> part_rowid_list = ibs.get_valid_part_rowids()
        >>> result = get_part_aids(ibs, part_rowid_list)
        >>> print(result)
    """
    aid_list = ibs.db.get(const.PART_TABLE, ('annot_rowid',), part_rowid_list, assume_unique=assume_unique)
    return aid_list


@register_ibs_method
def get_part_annot_rowids(ibs, part_rowid_list):
    return ibs.get_part_aids(part_rowid_list)


@register_ibs_method
def get_part_annot_uuids(ibs, part_rowid_list):
    aid_list = ibs.get_part_aids(part_rowid_list)
    annot_uuid_list = ibs.get_annot_uuids(aid_list)
    return annot_uuid_list


@register_ibs_method
@accessor_decors.getter_1to1
@register_api('/api/part/theta/', methods=['GET'])
def get_part_thetas(ibs, part_rowid_list):
    r"""
    Returns:
        theta_list (list): a list of floats describing the angles of each part

    CommandLine:
        python -m ibeis.control.manual_part_funcs --test-get_part_thetas

    RESTful:
        Method: GET
        URL:    /api/part/theta/

    Example:
        >>> # ENABLE_DOCTEST
        >>> from ibeis.control.manual_part_funcs import *  # NOQA
        >>> import ibeis
        >>> ibs = ibeis.opendb('NAUT_test')
        >>> part_rowid_list = ibs.get_valid_part_rowids()
        >>> result = get_part_thetas(ibs, part_rowid_list)
        >>> print(result)
        []
    """
    theta_list = ibs.db.get(const.PART_TABLE, ('part_theta',), part_rowid_list)
    return theta_list


@register_ibs_method
@accessor_decors.getter_1to1
@register_api('/api/part/uuid/', methods=['GET'])
def get_part_uuids(ibs, part_rowid_list):
    r"""
    Returns:
        list: part_uuid_list a list of part uuids by part_rowid

    RESTful:
        Method: GET
        URL:    /api/part/uuid/
    """
    part_uuid_list = ibs.db.get(const.PART_TABLE, ('part_uuid',), part_rowid_list)
    return part_uuid_list


@register_ibs_method
@accessor_decors.getter_1to1
# @register_api('/api/part/uuid/valid/', methods=['GET'])
def get_valid_part_uuids(ibs):
    r"""
    Returns:
        list: part_uuid_list a list of part uuids for all valid part_rowids
    """
    part_rowid_list = ibs.get_valid_part_rowids()
    part_uuid_list = ibs.get_part_uuids(part_rowid_list)
    return part_uuid_list


@register_ibs_method
@accessor_decors.getter_1to1
@register_api('/api/part/vert/', methods=['GET'])
def get_part_verts(ibs, part_rowid_list):
    r"""
    Returns:
        vert_list (list): the vertices that form the polygon of each part

    RESTful:
        Method: GET
        URL:    /api/part/vert/
    """
    from ibeis.algo.preproc import preproc_part
    vertstr_list = ibs.db.get(const.PART_TABLE, ('part_verts',), part_rowid_list)
    vert_list = preproc_part.postget_part_verts(vertstr_list)
    #vert_list = [eval(vertstr, {}, {}) for vertstr in vertstr_list]
    return vert_list


@register_ibs_method
@accessor_decors.getter_1to1
@register_api('/api/part/vert/rotated/', methods=['GET'])
def get_part_rotated_verts(ibs, part_rowid_list):
    r"""
    Returns:
        rotated_vert_list (list): verticies after rotation by theta.

    RESTful:
        Method: GET
        URL:    /api/part/vert/rotated/
    """
    import vtool_ibeis as vt
    vert_list = ibs.get_part_verts(part_rowid_list)
    theta_list = ibs.get_part_thetas(part_rowid_list)
    # Convex bounding boxes for verticies
    bbox_list = vt.geometry.bboxes_from_vert_list(vert_list)
    rot_list = [vt.rotation_around_bbox_mat3x3(theta, bbox)
                for theta, bbox in zip(theta_list, bbox_list)]
    rotated_vert_list = [vt.transform_points_with_homography(rot, np.array(verts).T).T.tolist()
                         for rot, verts in zip(rot_list, vert_list)]
    #vert_list = [eval(vertstr, {}, {}) for vertstr in vertstr_list]
    return rotated_vert_list


@register_ibs_method
@accessor_decors.getter_1to1
@register_api('/api/part/note/', methods=['GET'])
def get_part_notes(ibs, part_rowid_list):
    r"""
    Returns:
        part_notes_list (list): a list of part notes

    RESTful:
        Method: GET
        URL:    /api/part/note/
    """
    part_notes_list = ibs.db.get(const.PART_TABLE, (PART_NOTE,), part_rowid_list)
    return part_notes_list


@register_ibs_method
@accessor_decors.getter_1to1
@register_api('/api/part/type/', methods=['GET'])
def get_part_types(ibs, part_rowid_list):
    r"""
    Returns:
        part_notes_list (list): a list of part notes

    RESTful:
        Method: GET
        URL:    /api/part/note/
    """
    part_type_list = ibs.db.get(const.PART_TABLE, ('part_type',), part_rowid_list)
    return part_type_list


@register_ibs_method
@accessor_decors.getter_1to1
@register_api('/api/part/viewpoint/', methods=['GET'])
def get_part_viewpoints(ibs, part_rowid_list):
    r"""
    Returns:
        part_notes_list (list): a list of part notes

    RESTful:
        Method: GET
        URL:    /api/part/note/
    """
    part_viewpoint_list = ibs.db.get(const.PART_TABLE, ('part_viewpoint',), part_rowid_list)
    return part_viewpoint_list


@register_ibs_method
@accessor_decors.getter_1to1
@register_api('/api/part/num/vert/', methods=['GET'])
def get_part_num_verts(ibs, part_rowid_list):
    r"""
    Returns:
        nVerts_list (list): the number of vertices that form the polygon of each part

    RESTful:
        Method: GET
        URL:    /api/part/num/vert/
    """
    nVerts_list = ibs.db.get(const.PART_TABLE, (PART_NUM_VERTS,), part_rowid_list)
    return nVerts_list


@register_ibs_method
@accessor_decors.getter_1to1
@accessor_decors.cache_getter(const.PART_TABLE, PART_QUALITY)
@register_api('/api/part/quality/', methods=['GET'])
def get_part_qualities(ibs, part_rowid_list, eager=True):
    r"""
    part_quality_list <- part.part_quality[part_rowid_list]

    gets data from the "native" column "part_quality" in the "part" table

    Args:
        part_rowid_list (list):

    Returns:
        list: part_quality_list

    TemplateInfo:
        Tgetter_table_column
        col = part_quality
        tbl = part

    SeeAlso:
        ibeis.const.QUALITY_INT_TO_TEXT

    RESTful:
        Method: GET
        URL:    /api/part/quality/

    Example:
        >>> # ENABLE_DOCTEST
        >>> from ibeis.control.manual_part_funcs import *  # NOQA
        >>> ibs, qreq_ = testdata_ibs()
        >>> part_rowid_list = ibs._get_all_part_rowids()
        >>> eager = True
        >>> part_quality_list = ibs.get_part_qualities(part_rowid_list, eager=eager)
        >>> print('part_quality_list = %r' % (part_quality_list,))
        >>> assert len(part_rowid_list) == len(part_quality_list)
    """
    id_iter = part_rowid_list
    colnames = (PART_QUALITY,)
    part_quality_list = ibs.db.get(
        const.PART_TABLE, colnames, id_iter, id_colname='rowid', eager=eager)
    return part_quality_list


@register_ibs_method
@accessor_decors.getter_1to1
@register_api('/api/part/quality/text/', methods=['GET'])
def get_part_quality_texts(ibs, part_rowid_list):
    r"""
    Auto-docstr for 'get_part_quality_texts'

    RESTful:
        Method: GET
        URL:    /api/part/quality/text/
    """
    quality_list = ibs.get_part_qualities(part_rowid_list)
    quality_text_list = ut.dict_take(const.QUALITY_INT_TO_TEXT, quality_list)
    return quality_text_list


@register_ibs_method
@accessor_decors.getter_1to1
# @register_api('/api/part/isjunk/', methods=['GET'])
def get_part_isjunk(ibs, part_rowid_list):
    r"""
    Auto-docstr for 'get_part_isjunk'
    """
    qual_list = ibs.get_part_qualities(part_rowid_list)
    #isjunk_list = [qual == const.QUALITY_TEXT_TO_INT['junk'] for qual in qual_list]
    isjunk_list = [qual in const.QUALITY_TEXT_TO_INTS['junk'] for qual in qual_list]
    return isjunk_list


@register_ibs_method
@accessor_decors.getter_1to1
@register_api('/api/part/reviewed/', methods=['GET'])
def get_part_reviewed(ibs, part_rowid_list):
    r"""
    Returns:
        list_ (list): "All Instances Found" flag, true if all objects of interest
    (animals) have an PART in the part

    RESTful:
        Method: GET
        URL:    /api/part/reviewed/
    """
    reviewed_list = ibs.db.get(const.PART_TABLE, ('part_toggle_reviewed',), part_rowid_list)
    return reviewed_list


@register_ibs_method
@accessor_decors.getter_1to1
def get_part_tag_text(ibs, part_rowid_list, eager=True, nInput=None):
    r""" part_tags_list <- part.part_tags[part_rowid_list]

    gets data from the "native" column "part_tags" in the "part" table

    Args:
        part_rowid_list (list):

    Returns:
        list: part_tags_list

    Example:
        >>> # DISABLE_DOCTEST
        >>> from ibeis.control.manual_part_funcs import *  # NOQA
        >>> ibs, config2_ = testdata_ibs()
        >>> part_rowid_list = ibs._get_all_part_rowids()
        >>> eager = True
        >>> part_tags_list = ibs.get_part_tag_text(part_rowid_list, eager=eager)
        >>> assert len(part_rowid_list) == len(part_tags_list)
    """
    id_iter = part_rowid_list
    colnames = (PART_TAG_TEXT,)
    part_tags_list = ibs.db.get(
        const.PART_TABLE, colnames, id_iter, id_colname='rowid', eager=eager, nInput=nInput)
    return part_tags_list


@register_ibs_method
@accessor_decors.getter_1to1
@register_api('/api/part/staged/', methods=['GET'])
def get_part_staged_flags(ibs, part_rowid_list):
    r"""
    returns if an part is staged

    Args:
        ibs (IBEISController):  ibeis controller object
        part_rowid_list (int):  list of part ids

    Returns:
        list: part_staged_flag_list - True if part is staged

    CommandLine:
        python -m ibeis.control.manual_part_funcs --test-get_part_staged_flags

    RESTful:
        Method: GET
        URL:    /api/part/staged/

    Example:
        >>> # ENABLE_DOCTEST
        >>> from ibeis.control.manual_part_funcs import *  # NOQA
        >>> import ibeis
        >>> ibs = ibeis.opendb('testdb1')
        >>> part_rowid_list = ibs.get_valid_part_rowids()
        >>> gid_list = get_part_staged_flags(ibs, part_rowid_list)
        >>> result = str(gid_list)
        >>> print(result)
    """
    part_staged_flag_list = ibs.db.get(const.PART_TABLE,
                                          (PART_STAGED_FLAG,), part_rowid_list)
    return part_staged_flag_list


@register_ibs_method
@accessor_decors.getter_1to1
@register_api('/api/part/staged/uuid/', methods=['GET'])
def get_part_staged_uuids(ibs, aid_list):
    r"""
    Returns:
        list: part_uuid_list a list of image uuids by aid

    RESTful:
        Method: GET
        URL:    /api/part/staged/uuid/
    """
    part_uuid_list = ibs.db.get(const.PART_TABLE, ('part_staged_uuid',), aid_list)
    return part_uuid_list


@register_ibs_method
@accessor_decors.getter_1to1
@register_api('/api/part/staged/user/', methods=['GET'])
def get_part_staged_user_ids(ibs, part_rowid_list):
    r"""
    returns if an part is staged

    Args:
        ibs (IBEISController):  ibeis controller object
        part_rowid_list (int):  list of part ids

    Returns:
        list: part_staged_user_id_list - True if part is staged

    CommandLine:
        python -m ibeis.control.manual_part_funcs --test-get_part_staged_user_ids

    RESTful:
        Method: GET
        URL:    /api/part/staged/user/

    Example:
        >>> # ENABLE_DOCTEST
        >>> from ibeis.control.manual_part_funcs import *  # NOQA
        >>> import ibeis
        >>> ibs = ibeis.opendb('testdb1')
        >>> part_rowid_list = ibs.get_valid_part_rowids()
        >>> gid_list = get_part_staged_user_ids(ibs, part_rowid_list)
        >>> result = str(gid_list)
        >>> print(result)
    """
    part_staged_user_id_list = ibs.db.get(const.PART_TABLE,
                                           (PART_STAGED_USER_ID,), part_rowid_list)
    return part_staged_user_id_list


@register_ibs_method
@accessor_decors.getter_1to1
@register_api('/api/part/staged/metadata/', methods=['GET'])
def get_part_staged_metadata(ibs, part_rowid_list, return_raw=False):
    r"""
    Returns:
        list_ (list): part metadata dictionary

    RESTful:
        Method: GET
        URL:    /api/part/staged/metadata/
    """
    metadata_str_list = ibs.db.get(const.PART_TABLE, ('part_staged_metadata_json',), part_rowid_list)
    metadata_list = []
    for metadata_str in metadata_str_list:
        if metadata_str in [None, '']:
            metadata_dict = {}
        else:
            metadata_dict = metadata_str if return_raw else ut.from_json(metadata_str)
        metadata_list.append(metadata_dict)
    return metadata_list


#### SETTERS ####  # NOQA


@register_ibs_method
@accessor_decors.setter
@register_api('/api/part/annot/rowid/', methods=['PUT'])
def _set_part_aid(ibs, part_rowid_list, aid_list):
    r"""
    Sets part notes

    RESTful:
        Method: PUT
        URL:    /api/part/annot/rowid
    """
    id_iter = ((part_rowid,) for part_rowid in part_rowid_list)
    val_iter = ((confidence,) for confidence in aid_list)
    ibs.db.set(const.PART_TABLE, ('annot_rowid',), val_iter, id_iter)


@register_ibs_method
@accessor_decors.setter
@register_api('/api/part/bbox/', methods=['PUT'])
def set_part_bboxes(ibs, part_rowid_list, bbox_list):
    r"""
    Sets bboxes of a list of parts by part_rowid,

    Args:
        part_rowid_list (list of rowids): list of part rowids
        bbox_list (list of (x, y, w, h)): new bounding boxes for each part_rowid

    Note:
        set_part_bboxes is a proxy for set_part_verts

    RESTful:
        Method: PUT
        URL:    /api/part/bbox/
    """
    from vtool_ibeis import geometry
    # changing the bboxes also changes the bounding polygon
    vert_list = geometry.verts_list_from_bboxes_list(bbox_list)
    # naively overwrite the bounding polygon with a rectangle - for now trust the user!
    ibs.set_part_verts(part_rowid_list, vert_list)


@register_ibs_method
@accessor_decors.setter
@register_api('/api/part/detect/confidence/', methods=['PUT'])
def set_part_detect_confidence(ibs, part_rowid_list, confidence_list):
    r"""
    Sets part notes

    RESTful:
        Method: PUT
        URL:    /api/part/detect/confidence/
    """
    id_iter = ((part_rowid,) for part_rowid in part_rowid_list)
    val_iter = ((confidence,) for confidence in confidence_list)
    ibs.db.set(const.PART_TABLE, ('part_detect_confidence',), val_iter, id_iter)


@register_ibs_method
@accessor_decors.setter
@register_api('/api/part/note/', methods=['PUT'])
def set_part_notes(ibs, part_rowid_list, notes_list):
    r"""
    Sets part notes

    RESTful:
        Method: PUT
        URL:    /api/part/note/
    """
    id_iter = ((part_rowid,) for part_rowid in part_rowid_list)
    val_iter = ((notes,) for notes in notes_list)
    ibs.db.set(const.PART_TABLE, (PART_NOTE,), val_iter, id_iter)


@register_ibs_method
@accessor_decors.setter
@register_api('/api/part/type/', methods=['PUT'])
def set_part_types(ibs, part_rowid_list, type_list):
    r"""
    Sets part notes

    RESTful:
        Method: PUT
        URL:    /api/part/note/
    """
    id_iter = ((part_rowid,) for part_rowid in part_rowid_list)
    val_iter = ((type_,) for type_ in type_list)
    ibs.db.set(const.PART_TABLE, ('part_type',), val_iter, id_iter)


@register_ibs_method
@accessor_decors.setter
@register_api('/api/part/viewpoint/', methods=['PUT'])
def set_part_viewpoints(ibs, part_rowid_list, viewpoint_list):
    r"""
    Sets part notes

    RESTful:
        Method: PUT
        URL:    /api/part/note/
    """
    id_iter = ((part_rowid,) for part_rowid in part_rowid_list)
    val_iter = ((viewpoint_,) for viewpoint_ in viewpoint_list)
    ibs.db.set(const.PART_TABLE, ('part_viewpoint',), val_iter, id_iter)


@register_ibs_method
@accessor_decors.setter
def set_part_tag_text(ibs, part_rowid_list, part_tags_list, duplicate_behavior='error'):
    r""" part_tags_list -> part.part_tags[part_rowid_list]

    Args:
        part_rowid_list
        part_tags_list

    """
    #print('[ibs] set_part_tag_text of part_rowid_list=%r to tags=%r' % (part_rowid_list, part_tags_list))
    id_iter = part_rowid_list
    colnames = (PART_TAG_TEXT,)
    ibs.db.set(const.PART_TABLE, colnames, part_tags_list,
               id_iter, duplicate_behavior=duplicate_behavior)


@register_ibs_method
@accessor_decors.setter
@register_api('/api/part/theta/', methods=['PUT'])
def set_part_thetas(ibs, part_rowid_list, theta_list):
    r"""
    Sets thetas of a list of part_rowid_list

    RESTful:
        Method: PUT
        URL:    /api/part/theta/
    """
    id_iter = ((part_rowid,) for part_rowid in part_rowid_list)
    val_list = ((theta,) for theta in theta_list)
    ibs.db.set(const.PART_TABLE, (PART_THETA,), val_list, id_iter)


@register_ibs_method
@accessor_decors.setter
@register_api('/api/part/vert/', methods=['PUT'])
def set_part_verts(ibs, part_rowid_list, verts_list):
    r"""
    Sets the vertices [(x, y), ...] of a list of part_rowid_list

    RESTful:
        Method: PUT
        URL:    /api/part/vert/
    """
    from vtool_ibeis import geometry
    nInput = len(part_rowid_list)
    # Compute data to set
    if isinstance(verts_list, np.ndarray):
        verts_list = verts_list.tolist()
    for index, vert_list in enumerate(verts_list):
        if isinstance(vert_list, np.ndarray):
            verts_list[index] = vert_list.tolist()
    num_verts_list   = list(map(len, verts_list))
    verts_as_strings = list(map(six.text_type, verts_list))
    id_iter1 = ((part_rowid,) for part_rowid in part_rowid_list)
    # also need to set the internal number of vertices
    val_iter1 = ((num_verts, verts) for (num_verts, verts)
                 in zip(num_verts_list, verts_as_strings))
    colnames = (PART_NUM_VERTS, PART_VERTS,)
    # SET VERTS in PART_TABLE
    ibs.db.set(const.PART_TABLE, colnames, val_iter1, id_iter1, nInput=nInput)
    # changing the vertices also changes the bounding boxes
    bbox_list = geometry.bboxes_from_vert_list(verts_list)      # new bboxes
    xtl_list, ytl_list, width_list, height_list = list(zip(*bbox_list))
    val_iter2 = zip(xtl_list, ytl_list, width_list, height_list)
    id_iter2 = ((part_rowid,) for part_rowid in part_rowid_list)
    colnames = ('part_xtl', 'part_ytl', 'part_width', 'part_height',)
    # SET BBOX in PART_TABLE
    ibs.db.set(const.PART_TABLE, colnames, val_iter2, id_iter2, nInput=nInput)


@register_ibs_method
@accessor_decors.setter
@accessor_decors.cache_invalidator(const.PART_TABLE, [PART_QUALITY], rowidx=0)
@register_api('/api/part/quality/', methods=['PUT'])
def set_part_qualities(ibs, part_rowid_list, part_quality_list):
    r"""
    part_quality_list -> part.part_quality[part_rowid_list]

    A quality is an integer representing the following types:

    Args:
        part_rowid_list
        part_quality_list

    SeeAlso:
        ibeis.const.QUALITY_INT_TO_TEXT

    RESTful:
        Method: PUT
        URL:    /api/part/quality/
    """
    id_iter = part_rowid_list
    colnames = (PART_QUALITY,)
    ibs.db.set(const.PART_TABLE, colnames, part_quality_list, id_iter)


@register_ibs_method
@register_api('/api/part/quality/text/', methods=['PUT'])
def set_part_quality_texts(ibs, part_rowid_list, quality_text_list):
    r"""
    Auto-docstr for 'set_part_quality_texts'

    RESTful:
        Method: PUT
        URL:    /api/part/quality/text/
    """
    if not ut.isiterable(part_rowid_list):
        part_rowid_list = [part_rowid_list]
    if isinstance(quality_text_list, six.string_types):
        quality_text_list = [quality_text_list]
    quality_list = ut.dict_take(const.QUALITY_TEXT_TO_INT, quality_text_list)
    ibs.set_part_qualities(part_rowid_list, quality_list)


@register_ibs_method
@accessor_decors.setter
@register_api('/api/part/reviewed/', methods=['PUT'])
def set_part_reviewed(ibs, part_rowid_list, reviewed_list):
    r"""
    Sets the part all instances found bit

    RESTful:
        Method: PUT
        URL:    /api/part/reviewed/
    """
    id_iter = ((part_rowid,) for part_rowid in part_rowid_list)
    val_list = ((reviewed,) for reviewed in reviewed_list)
    ibs.db.set(const.PART_TABLE, ('part_toggle_reviewed',), val_list, id_iter)


@register_ibs_method
@accessor_decors.setter
@accessor_decors.cache_invalidator(const.PART_TABLE, [PART_STAGED_FLAG], rowidx=0)
@register_api('/api/part/staged/', methods=['PUT'])
def _set_part_staged_flags(ibs, part_rowid_list, flag_list):
    r"""
    Sets if an part is staged

    RESTful:
        Method: PUT
        URL:    /api/part/staged/
    """
    id_iter = ((part_rowid,) for part_rowid in part_rowid_list)
    val_iter = ((flag,) for flag in flag_list)
    ibs.db.set(const.PART_TABLE, (PART_STAGED_FLAG,), val_iter, id_iter)


@register_ibs_method
@register_api('/api/part/staged/uuid/', methods=['PUT'])
def set_part_staged_uuids(ibs, aid_list, part_uuid_list):
    r"""
    Returns:
        list_ (list): all nids of known animals
        (does not include unknown names)
    """
    id_iter = ((aid,) for aid in aid_list)
    val_iter = ((part_uuid,) for part_uuid in part_uuid_list)
    ibs.db.set(const.PART_TABLE, ('part_staged_uuid',), val_iter, id_iter)
    flag_list = [
        part_uuid is not None
        for part_uuid in part_uuid_list
    ]
    ibs._set_part_staged_flags(aid_list, flag_list)


@register_ibs_method
@accessor_decors.setter
@accessor_decors.cache_invalidator(const.PART_TABLE, [PART_STAGED_USER_ID], rowidx=0)
@register_api('/api/part/staged/user/', methods=['PUT'])
def set_part_staged_user_ids(ibs, part_rowid_list, user_id_list):
    r"""
    Sets the staged part user id

    RESTful:
        Method: PUT
        URL:    /api/part/staged/user/
    """
    id_iter = ((part_rowid,) for part_rowid in part_rowid_list)
    val_iter = ((user_id,) for user_id in user_id_list)
    ibs.db.set(const.PART_TABLE, (PART_STAGED_USER_ID,), val_iter, id_iter)


@register_ibs_method
@accessor_decors.setter
@register_api('/api/part/staged/metadata/', methods=['PUT'])
def set_part_staged_metadata(ibs, part_rowid_list, metadata_dict_list):
    r"""
    Sets the part's staged metadata using a metadata dictionary

    RESTful:
        Method: PUT
        URL:    /api/part/staged/metadata/

    CommandLine:
        python -m ibeis.control.manual_part_funcs --test-set_part_staged_metadata

    Example:
        >>> # ENABLE_DOCTEST
        >>> from ibeis.control.manual_part_funcs import *  # NOQA
        >>> import ibeis
        >>> import random
        >>> # build test data
        >>> ibs = ibeis.opendb('testdb1')
        >>> aid_list = ibs.get_valid_aids()[0:1]
        >>> bbox_list = [[0, 0, 100, 100]] * len(aid_list)
        >>> part_rowid_list = ibs,add_parts(aid_list, bbox_list=bbox_list)
        >>> metadata_dict_list = [
        >>>     {'test': random.uniform(0.0, 1.0)},
        >>> ] * len(part_rowid_list)
        >>> print(ut.repr2(metadata_dict_list))
        >>> ibs.set_part_staged_metadata(part_rowid_list, metadata_dict_list)
        >>> # verify results
        >>> metadata_dict_list_ = ibs.get_part_staged_metadata(part_rowid_list)
        >>> print(ut.repr2(metadata_dict_list_))
        >>> assert metadata_dict_list == metadata_dict_list_
        >>> metadata_str_list = [ut.to_json(metadata_dict) for metadata_dict in metadata_dict_list]
        >>> print(ut.repr2(metadata_str_list))
        >>> metadata_str_list_ = ibs.get_part_staged_metadata(part_rowid_list, return_raw=True)
        >>> print(ut.repr2(metadata_str_list_))
        >>> assert metadata_str_list == metadata_str_list_
        >>> ibs.delete_parts(part_rowid_list)
    """
    id_iter = ((part_rowid,) for part_rowid in part_rowid_list)
    metadata_str_list = []
    for metadata_dict in metadata_dict_list:
        metadata_str = ut.to_json(metadata_dict)
        metadata_str_list.append(metadata_str)
    val_list = ((metadata_str,) for metadata_str in metadata_str_list)
    ibs.db.set(const.PART_TABLE, ('part_staged_metadata_json',), val_list, id_iter)


@register_ibs_method
@accessor_decors.getter_1to1
@register_api('/api/part/metadata/', methods=['GET'])
def get_part_metadata(ibs, part_rowid_list, return_raw=False):
    r"""
    Returns:
        list_ (list): part metadata dictionary

    RESTful:
        Method: GET
        URL:    /api/part/metadata/
    """
    metadata_str_list = ibs.db.get(const.PART_TABLE, ('part_metadata_json',), part_rowid_list)
    metadata_list = []
    for metadata_str in metadata_str_list:
        if metadata_str in [None, '']:
            metadata_dict = {}
        else:
            metadata_dict = metadata_str if return_raw else ut.from_json(metadata_str)
        metadata_list.append(metadata_dict)
    return metadata_list


@register_ibs_method
@accessor_decors.setter
@register_api('/api/part/metadata/', methods=['PUT'])
def set_part_metadata(ibs, part_rowid_list, metadata_dict_list):
    r"""
    Sets the part's metadata using a metadata dictionary

    RESTful:
        Method: PUT
        URL:    /api/part/metadata/

    CommandLine:
        python -m ibeis.control.manual_part_funcs --test-set_part_metadata

    Example:
        >>> # ENABLE_DOCTEST
        >>> from ibeis.control.manual_part_funcs import *  # NOQA
        >>> import ibeis
        >>> import random
        >>> # build test data
        >>> ibs = ibeis.opendb('testdb1')
        >>> aid_list = ibs.get_valid_aids()[0:1]
        >>> bbox_list = [[0, 0, 100, 100]] * len(aid_list)
        >>> part_rowid_list = ibs.add_parts(aid_list, bbox_list=bbox_list)
        >>> metadata_dict_list = [
        >>>     {'test': random.uniform(0.0, 1.0)},
        >>> ]
        >>> print(ut.repr2(metadata_dict_list))
        >>> ibs.set_part_metadata(part_rowid_list, metadata_dict_list)
        >>> # verify results
        >>> metadata_dict_list_ = ibs.get_part_metadata(part_rowid_list)
        >>> print(ut.repr2(metadata_dict_list_))
        >>> assert metadata_dict_list == metadata_dict_list_
        >>> metadata_str_list = [ut.to_json(metadata_dict) for metadata_dict in metadata_dict_list]
        >>> print(ut.repr2(metadata_str_list))
        >>> metadata_str_list_ = ibs.get_part_metadata(part_rowid_list, return_raw=True)
        >>> print(ut.repr2(metadata_str_list_))
        >>> assert metadata_str_list == metadata_str_list_
        >>> ibs.delete_parts(part_rowid_list)
    """
    id_iter = ((part_rowid,) for part_rowid in part_rowid_list)
    metadata_str_list = []
    for metadata_dict in metadata_dict_list:
        metadata_str = ut.to_json(metadata_dict)
        metadata_str_list.append(metadata_str)
    val_list = ((metadata_str,) for metadata_str in metadata_str_list)
    ibs.db.set(const.PART_TABLE, ('part_metadata_json',), val_list, id_iter)


@register_ibs_method
@accessor_decors.getter_1to1
@register_api('/api/part/contour/', methods=['GET'])
def get_part_contour(ibs, part_rowid_list, return_raw=False):
    r"""
    Returns:
        list_ (list): part contour dictionary

    RESTful:
        Method: GET
        URL:    /api/part/contour/
    """
    contour_str_list = ibs.db.get(const.PART_TABLE, ('part_contour_json',), part_rowid_list)
    contour_list = []
    for contour_str in contour_str_list:
        if contour_str in [None, '']:
            contour_dict = {}
        else:
            contour_dict = contour_str if return_raw else ut.from_json(contour_str)
        contour_list.append(contour_dict)
    return contour_list


@register_ibs_method
@accessor_decors.setter
@register_api('/api/part/contour/', methods=['PUT'])
def set_part_contour(ibs, part_rowid_list, contour_dict_list):
    r"""
    Sets the part's contour using a contour dictionary

    RESTful:
        Method: PUT
        URL:    /api/part/contour/

    CommandLine:
        python -m ibeis.control.manual_part_funcs --test-set_part_contour

    Example:
        >>> # ENABLE_DOCTEST
        >>> from ibeis.control.manual_part_funcs import *  # NOQA
        >>> import ibeis
        >>> import random
        >>> # build test data
        >>> ibs = ibeis.opendb('testdb1')
        >>> aid_list = ibs.get_valid_aids()[0:1]
        >>> bbox_list = [[0, 0, 100, 100]] * len(aid_list)
        >>> part_rowid_list = ibs.add_parts(aid_list, bbox_list=bbox_list)
        >>> contour_dict_list = [
        >>>     {'test': random.uniform(0.0, 1.0)},
        >>> ]
        >>> print(ut.repr2(contour_dict_list))
        >>> ibs.set_part_contour(part_rowid_list, contour_dict_list)
        >>> # verify results
        >>> contour_dict_list_ = ibs.get_part_contour(part_rowid_list)
        >>> print(ut.repr2(contour_dict_list_))
        >>> assert contour_dict_list == contour_dict_list_
        >>> contour_str_list = [ut.to_json(contour_dict) for contour_dict in contour_dict_list]
        >>> print(ut.repr2(contour_str_list))
        >>> contour_str_list_ = ibs.get_part_contour(part_rowid_list, return_raw=True)
        >>> print(ut.repr2(contour_str_list_))
        >>> assert contour_str_list == contour_str_list_
        >>> ibs.delete_parts(part_rowid_list)
    """
    id_iter = ((part_rowid,) for part_rowid in part_rowid_list)
    contour_str_list = []
    for contour_dict in contour_dict_list:
        contour_str = ut.to_json(contour_dict)
        contour_str_list.append(contour_str)
    val_list = ((contour_str,) for contour_str in contour_str_list)
    ibs.db.set(const.PART_TABLE, ('part_contour_json',), val_list, id_iter)


#==========
# Testdata
#==========


def testdata_ibs():
    r"""
    Auto-docstr for 'testdata_ibs'
    """
    import ibeis
    ibs = ibeis.opendb('testdb1')
    qreq_ = None
    return ibs, qreq_


if __name__ == '__main__':
    r"""
    CommandLine:
        python -m ibeis.control.manual_part_funcs
        python -m ibeis.control.manual_part_funcs --allexamples
        python -m ibeis.control.manual_part_funcs --allexamples --noface --nosrc
    """
    import multiprocessing
    multiprocessing.freeze_support()  # for win32
    import utool as ut  # NOQA
    ut.doctest_funcs()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Very useful script to ensure you have all the modules you need

CommandLine:
    python -m ibeis.tests.assert_modules
    python -m ibeis.tests.assert_modules --test-assert_modules --nolibdep
    python -m ibeis.tests.assert_modules --test-assert_modules


MacFix:
    # Remove the copy of pyrf
    /opt/local/Library/Frameworks/Python.framework/Versions/2.7/lib/python2.7/site-packages/
    sudo rm -rf pyrf-1.0.0.dev1-py2.7.egg/

Updater For Linux:
    sudo pip install matplotlib --upgrade
    sudo pip install Pillow --upgrade
    sudo pip install scipy --upgrade
    sudo pip install numpy --upgrade

    ./super_setup.py --bootstrap --upgrade
"""
from __future__ import absolute_import, division, print_function
import sys
import functools
import utool as ut
from pkg_resources import parse_version
from utool._internal.meta_util_six import get_funcname

ASSERT_FUNCS = []


#LIB_DEP = ut.get_argflag('--nolibdep')
LIB_DEP = not ut.get_argflag('--libdep')


def get_site_package_directories():
    import site
    import sys
    import six
    sitepackages = site.getsitepackages()
    if sys.platform.startswith('darwin'):
        if six.PY2:
            macports_site = '/opt/local/Library/Frameworks/Python.framework/Versions/2.7/lib/python2.7/site-packages'
        else:
            macports_site = '/opt/local/Library/Frameworks/Python.framework/Versions/3.5/lib/python3.5/site-packages'
            assert six.PY2, 'fix this for python 3'
        sitepackages = [macports_site] + sitepackages
    return sitepackages


def check_alternate_installs():
    '/Library/Python/2.7/site-packages'

    pass


def version_ge_target(version, target=None):
    if target is None:
        passed = True
    elif version is None:
        passed = False
    else:
        _version = version.replace('.dev1', '')
        passed = parse_version(_version) >= parse_version(target)
    return passed


def checkinfo(target=None, pipname=None):
    """
    checkinfo functions return info_dict containing __version__
    """
    def wrapper1(func):
        """
        Returns;
            tuple : passed, current_version, target, infodict, statustext, suggested_fix
        """

        # Decorator which adds funcs to ASSERT_FUNCS
        global ASSERT_FUNCS
        @functools.wraps(func)
        def checkinfo_wrapper(*args, **kwargs):
            suggested_fix = ''
            funcname = get_funcname(func)
            packagename = funcname.replace('_version', '')
            pipname_ = pipname if pipname is not None else packagename
            try:
                infodict = func(*args, **kwargs)
            except ImportError as ex:
                infodict = module_stdinfo_dict(None, name=pipname_)
                suggested_fix = 'pip install ' + pipname_
                if not sys.platform.startswith('win32'):
                    suggested_fix = 'sudo ' + suggested_fix
                return False, 'None', target, infodict, ut.formatex(ex), suggested_fix
            except Exception as ex:
                infodict = module_stdinfo_dict(None, name=pipname_)
                return False, 'None', target, infodict, ut.formatex(ex), 'Some unknown error in ' + packagename
            current_version = infodict['__version__']
            # Build status text
            msg = ut.repr2(infodict, strvals=True)
            msg += '\n' + '%s: %r >= (target=%r)?' % (funcname, current_version, target)
            statustext = ut.msgblock(infodict['__name__'], msg)
            # Check if passed
            passed = version_ge_target(current_version, target)
            # Suggest possible fix
            if not passed:
                suggested_fix = 'pip install ' + infodict['__name__'] + ' --upgrade'
                if not sys.platform.startswith('win32'):
                    suggested_fix = 'sudo ' + suggested_fix
            return passed, current_version, target, infodict, statustext, suggested_fix
        ASSERT_FUNCS.append(checkinfo_wrapper)
        return checkinfo_wrapper
    return wrapper1


def module_stdinfo_dict(module, versionattr='__version__', version=None, libdep=None, name=None, **kwargs):
    #if module is None:
    #    module = object
    infodict = {
        '__version__': version if module is None or version is not None else getattr(module, versionattr, None),
        '__name__': name if module is None else getattr(module, '__name__', name),
        '__file__': 'None' if module is None else getattr(module, '__file__', None),
    }
    if libdep is not None:
        infodict['libdep'] = libdep
    if not ut.QUIET:
        infodict.update(kwargs)
    return infodict


def reg_std_version_check(version, modname):
    def checkstdmod_version():
        mod = ut.import_modname(modname)
        return module_stdinfo_dict(mod)
    ut.set_funcname(checkstdmod_version, '%s_version' % (modname,))
    func = checkinfo(version)(checkstdmod_version)
    return func


reg_std_version_check('1.5.3', 'pynmea2')
#reg_std_version_check(None, 'wget')
# reg_std_version_check(None, 'pygco')
reg_std_version_check('6.0.8', 'pip')
reg_std_version_check('1.1.1', 'utool')
reg_std_version_check('0.12.3', 'skimage')
reg_std_version_check('1.1.6', 'cachetools')


# @checkinfo('1.1.1')
# def utool_version():
#     import utool
#     return module_stdinfo_dict(utool)

# @checkinfo('1.5.3')
# def pynmea2_version():
#     # for web
#     import pynmea2
#     return module_stdinfo_dict(pynmea2)

# @checkinfo('6.0.8')
# def pip_version():
#     import pip
#     return module_stdinfo_dict(pip)


@checkinfo(None)
def pyflann_version():
    from vtool_ibeis._pyflann_backend import pyflann as pyflann
    if LIB_DEP:
        libdep = None
    else:
        libdep = ut.get_dynlib_dependencies(pyflann.flannlib._name)
    return module_stdinfo_dict(pyflann, libdep=libdep)


@checkinfo('1.1.1')
def pyhesaff_version():
    import pyhesaff
    if LIB_DEP:
        libdep = None
    else:
        libdep = ut.get_dynlib_dependencies(pyhesaff.__LIB_FPATH__)
    return module_stdinfo_dict(pyhesaff, libdep=libdep)


@checkinfo('1.0.0')
def pyrf_version():
    import pyrf
    if LIB_DEP:
        libdep = None
    else:
        libdep = ut.get_dynlib_dependencies(pyrf.RF_CLIB._name)
    return module_stdinfo_dict(pyrf, libdep=libdep)


@checkinfo('1.0.1')
def vtool_version():
    import vtool_ibeis
    libdep = None
    if LIB_DEP:
        libdep = None
    else:
        try:
            libdep = ut.get_dynlib_dependencies(vtool_ibeis.sver_c_wrapper.lib_fname)
        except Exception:
            pass
    return module_stdinfo_dict(vtool_ibeis, libdep=libdep)


#@checkinfo('1.1.7')
#@checkinfo('2.4.0')
@checkinfo('3.1.0')
def pillow_version():
    from PIL import Image
    import PIL
    pil_path = PIL.__path__
    if len(PIL.__path__) > 1:
        print('WARNING!!! THERE ARE MULTIPLE PILS! %r ' % PIL.__path__)
    return module_stdinfo_dict(
        Image, versionattr='PILLOW_VERSION', image_version=Image.VERSION, pil_path=pil_path)


#@checkinfo('1.3.1')
@checkinfo('1.5.1')
def matplotlib_version():
    import matplotlib as mpl
    return module_stdinfo_dict(mpl)


@checkinfo('2.4.8')
def opencv_version():
    import cv2
    #print(cv2.getBuildInformation())
    if LIB_DEP:
        libdep = None
    else:
        libdep = ut.get_dynlib_dependencies(cv2.__file__)
    return module_stdinfo_dict(cv2, libdep=libdep)


@checkinfo('0.13.2')
def scipy_version():
    import scipy
    return module_stdinfo_dict(scipy)


@checkinfo('0.4.9')
def scipy_linalg_version():
    import scipy.linalg
    return module_stdinfo_dict(scipy.linalg)


@checkinfo('1.8.0')
def numpy_version():
    import numpy
    return module_stdinfo_dict(numpy)


#@checkinfo()
def theano_version():
    import theano
    return module_stdinfo_dict(theano)


#@checkinfo()
def lasagne_version():
    import lasagne
    return module_stdinfo_dict(lasagne)


@checkinfo('4.9.1')  # 4.10.1 on windows
def PyQt4_version():
    # FIXME, pyqt5 is also ok
    from PyQt4 import QtCore
    return module_stdinfo_dict(QtCore, 'PYQT_VERSION_STR')


@checkinfo('0.15.1')
def pandas_version():
    import pandas
    return module_stdinfo_dict(pandas)


@checkinfo('0.6.1')
def statsmodels_version():
    import statsmodels
    version = statsmodels.version.version
    return module_stdinfo_dict(statsmodels, version=version)


@checkinfo('0.10.1')
def flask_version():
    import flask
    return module_stdinfo_dict(flask)


@checkinfo('2.0.1')
def flask_cors_version():
    #import flask.ext.cors as flask_cors
    import flask_cors
    return module_stdinfo_dict(flask_cors)


@checkinfo('4.0.2')
def tornado_version():
    import tornado
    return module_stdinfo_dict(tornado, 'version')


@checkinfo()
def pygraphviz_version():
    import pygraphviz
    return module_stdinfo_dict(pygraphviz)


@checkinfo(None)
def networkx_version():
    # for web
    import networkx
    return module_stdinfo_dict(networkx)


def check_modules_exists():
    # Modules in this list don't really need to be inspected
    # just make sure they are there
    modname_list = [
        'simplejson',
        'flask',
        'parse',
        'tornado',
        'pandas',
        'statsmodels',
    ]
    failed_list = []
    for modname in modname_list:
        try:
            globals_ = {}
            locals_ = {}
            exec('import ' + modname, globals_, locals_)
        except ImportError:
            failed_list.append(modname)
        except Exception as ex:
            ut.printex(ex, 'Some othere error happened when importing %r ' % (modname,), iswarning=True)
            failed_list.append(modname)
    if len(failed_list) > 0:
        print('The following modules are not installed')
        print('\n'.join(failed_list))
    return ''


def assert_modules():
    """
    checkinfo functions return info_dict
    checkinfo_func

    CommandLine:
        python -m ibeis.tests.assert_modules --test-assert_modules

    Example:
        >>> # DOCTEST_ENABLE
        >>> from ibeis.tests.assert_modules import *   # NOQA
        >>> detailed_msg = assert_modules()
        >>> print(detailed_msg)
    """

    MACHINE_NAME = ut.get_computer_name()

    machine_info_lines = []

    machine_info_lines.append('sys.version = %r ' % (sys.version))
    machine_info_lines.append('PATH = ' + ut.repr2(ut.get_path_dirs()))
    machine_info_lines.append('\n\n\n============================')
    machine_info_lines.append('Begining assert modules main')
    machine_info_lines.append('* MACHINE_NAME = %r' % MACHINE_NAME)
    machine_info_text = '\n'.join(machine_info_lines)
    print(machine_info_text)

    statustext_list = []
    failed_list = []
    fix_list = []

    SHOW_STATUS = not ut.get_argflag(('--nostatus', '--nostat'))

    for checkinfo_wrapper in ASSERT_FUNCS:
        passed, current_version, target, infodict, statustext, suggested_fix = checkinfo_wrapper()
        funcname = get_funcname(checkinfo_wrapper)
        if SHOW_STATUS:
            statustext_list.append(statustext)
        if passed:
            statustext_list.append(funcname + ' ' + str(infodict['__version__']) + ' passed')
            #statustext_list.append('')
        else:
            failed_list.append(funcname + ' FAILED!!!')
            fix_list.append(suggested_fix)
            statustext_list.append(funcname + ' FAILED!!!')
        if SHOW_STATUS:
            statustext_list.append('')

    output_text = '\n'.join(statustext_list)

    failed_text = '\n'.join(failed_list)
    print(output_text)
    print(failed_text)
    check_exist_text = check_modules_exists()
    print(check_exist_text)
    fix_text = ''
    if len(fix_list) > 0:
        fix_text += ('suggested fixes:\n')
        fix_text += ('\n'.join(fix_list) + '\n')
        print(fix_text)

    detailed_msg = '\n'.join([
        machine_info_text,
        output_text,
        failed_text,
        check_exist_text,
        fix_text,
    ])

    return detailed_msg


if __name__ == '__main__':
    """
    CommandLine:
        python -m ibeis.tests.assert_modules

        python -c "import utool, ibeis.tests.assert_modules; utool.doctest_funcs(ibeis.tests.assert_modules, allexamples=True)"
        python -m ibeis.tests.assert_modules --allexamples
        python ~/code/ibeis/ibeis/tests/assert_modules.py

        python -m ibeis.tests.assert_modules
    """
    import multiprocessing
    multiprocessing.freeze_support()  # for win32
    import utool as ut  # NOQA
    print('in assert module main')
    #if len(sys.argv) == 1:
    if not any(argv.startswith('--test-') or argv.startswith('--exec') for argv in sys.argv):
        assert_modules()
    else:
        ut.doctest_funcs()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Runs IBIES gui
"""
from __future__ import absolute_import, division, print_function
import multiprocessing
import utool as ut
import ibeis  # NOQA
import sys

CMD = ut.get_argflag('--cmd')


def dependencies_for_myprogram():
    """ Let pyintaller find these modules

    References:
        http://stackoverflow.com/questions/18596410/importerror-no-module-named-mpl-toolkits-with-maptlotlib-1-3-0-and-py2exe
    """
    from guitool_ibeis.__PYQT__ import QtCore, QtGui  # Pyinstaller hacks  # NOQA
    from PyQt4 import QtCore, QtGui  # NOQA
    #from PyQt4 import QtCore, QtGui  # NOQA
    from scipy.sparse.csgraph import _validation  # NOQA
    from scipy.special import _ufuncs_cxx  # NOQA
    from mpl_toolkits.axes_grid1 import make_axes_locatable  # NOQA
    #import lru  # NOQA
    # Workaround for mpl_toolkits
    import importlib
    importlib.import_module('mpl_toolkits').__path__


def run_ibeis():
    r"""
    CommandLine:
        python -m ibeis
        python -m ibeis find_installed_tomcat
        python -m ibeis get_annot_groundtruth:1
    """
    #ut.set_process_title('IBEIS_main')
    #main_locals = ibeis.main()
    #ibeis.main_loop(main_locals)
    #ut.set_process_title('IBEIS_main')
    cmdline_varags = ut.get_cmdline_varargs()
    if len(cmdline_varags) > 0 and cmdline_varags[0] == 'rsync':
        from ibeis.scripts import rsync_ibeisdb
        rsync_ibeisdb.rsync_ibsdb_main()
        sys.exit(0)

    if ut.get_argflag('--devcmd'):
        # Hack to let devs mess around when using an installer version
        # TODO: add more hacks
        #import utool.tests.run_tests
        #utool.tests.run_tests.run_tests()
        ut.embed()
    # Run the tests of other modules
    elif ut.get_argflag('--run-utool-tests'):
        raise Exception('Deprecated functionality')
    elif ut.get_argflag('--run-vtool_ibeis-tests'):
        raise Exception('Deprecated functionality')
    elif ut.get_argflag(('--run-ibeis-tests', '--run-tests')):
        raise Exception('Deprecated functionality')

    if ut.get_argflag('-e'):
        """
        ibeis -e print -a default -t default
        """
        # Run dev script if -e given
        import ibeis.dev  # NOQA
        ibeis.dev.devmain()
        print('... exiting')
        sys.exit(0)

    # Attempt to run a test using the funciton name alone
    # with the --tf flag
    import ibeis.tests.run_tests
    import ibeis.tests.reset_testdbs
    import ibeis.scripts.thesis
    ignore_prefix = [
        #'ibeis.tests',
        'ibeis.control.__SQLITE3__',
        '_autogen_explicit_controller']
    ignore_suffix = ['_grave']
    func_to_module_dict = {
        'demo_bayesnet': 'ibeis.unstable.demobayes',
    }
    ut.main_function_tester('ibeis', ignore_prefix, ignore_suffix,
                            func_to_module_dict=func_to_module_dict)

    #if ut.get_argflag('-e'):
    #    import ibeis
    #    expt_kw = ut.get_arg_dict(ut.get_func_kwargs(ibeis.run_experiment),
    #    prefix_list=['--', '-'])
    #    ibeis.run_experiment(**expt_kw)
    #    sys.exit(0)

    doctest_modname = ut.get_argval(
        ('--doctest-module', '--tmod', '-tm', '--testmod'),
        type_=str, default=None, help_='specify a module to doctest')
    if doctest_modname is not None:
        """
        Allow any doctest to be run the main ibeis script

        python -m ibeis --tmod utool.util_str --test-align:0
        python -m ibeis --tmod ibeis.algo.hots.pipeline --test-request_ibeis_query_L0:0 --show
        python -m ibeis --tf request_ibeis_query_L0:0 --show
        ./dist/ibeis/IBEISApp --tmod ibeis.algo.hots.pipeline --test-request_ibeis_query_L0:0 --show  # NOQA
        ./dist/ibeis/IBEISApp --tmod utool.util_str --test-align:0
        ./dist/IBEIS.app/Contents/MacOS/IBEISApp --tmod utool.util_str --test-align:0
        ./dist/IBEIS.app/Contents/MacOS/IBEISApp --run-utool-tests
        ./dist/IBEIS.app/Contents/MacOS/IBEISApp --run-vtool_ibeis-tests
        """
        print('[ibeis] Testing module')
        mod_alias_list = {
            'exptdraw': 'ibeis.expt.experiment_drawing'
        }
        doctest_modname = mod_alias_list.get(doctest_modname, doctest_modname)
        module = ut.import_modname(doctest_modname)
        (nPass, nTotal, failed_list, error_report_list) = ut.doctest_funcs(module=module)
        retcode = 1 - (len(failed_list) == 0)
        #print(module)
        sys.exit(retcode)

    import ibeis
    main_locals = ibeis.main()
    execstr = ibeis.main_loop(main_locals)
    # <DEBUG CODE>
    if 'back' in main_locals and CMD:
        back = main_locals['back']
        front = getattr(back, 'front', None)  # NOQA
        #front = back.front
        #ui = front.ui
    ibs = main_locals['ibs']  # NOQA
    exec(execstr)
    # </DEBUG CODE>


if __name__ == '__main__':
    multiprocessing.freeze_support()  # for win32
    run_ibeis()



# -*- coding: utf-8 -*-
### __init__.py ###
# flake8: noqa
from __future__ import absolute_import, division, print_function

import utool as ut
ut.noinject(__name__, '[ibeis.control.__init__]', DEBUG=False)


from ibeis.control import DB_SCHEMA
from ibeis.control import IBEISControl
from ibeis.control import _sql_helpers
from ibeis.control import accessor_decors
import utool
print, rrr, profile = utool.inject2(__name__)


def reload_subs(verbose=True):
    """ Reloads ibeis.control and submodules """
    rrr(verbose=verbose)
    getattr(DB_SCHEMA, 'rrr', lambda verbose: None)(verbose=verbose)
    getattr(IBEISControl, 'rrr', lambda verbose: None)(verbose=verbose)
    getattr(_sql_helpers, 'rrr', lambda verbose: None)(verbose=verbose)
    getattr(accessor_decors, 'rrr', lambda verbose: None)(verbose=verbose)
    rrr(verbose=verbose)
rrrr = reload_subs

IMPORT_TUPLES = [
    ('DB_SCHEMA', None, False),
    ('IBEISControl', None, False),
    ('_sql_helpers', None, False),
    ('accessor_decors', None, False),
]
"""
Regen Command:
    cd /home/joncrall/code/ibeis/ibeis/control
    makeinit.py -x DBCACHE_SCHEMA_CURRENT DB_SCHEMA_CURRENT _grave_template manual_ibeiscontrol_funcs template_definitions templates _autogen_ibeiscontrol_funcs
"""
